import numpy as np
import matplotlib.pyplot as plt

def plotmesh(mesh, **kwargs):
    ax = plt
    if 'ax' in kwargs: ax = kwargs.pop('ax')
    x, lines = mesh.points[:, 0], mesh.simplices
    ax.plot(x, np.zeros_like(x), "x-")

def meshWithBoundaries(x, lines, **kwargs):
    ax = plt
    if 'ax' in kwargs: ax = kwargs.pop('ax')
    ax.plot(x, np.zeros_like(x), "x-")

def meshWithData(**kwargs):
    ax = plt
    if 'ax' in kwargs: ax = kwargs.pop('ax')
    x = kwargs.pop('x')
    title, suptitle = None, None
    if 'title' in kwargs: title = kwargs['title']
    if 'suptitle' in kwargs: suptitle = kwargs['suptitle']
    point_data, cell_data = None, None
    if 'data' in kwargs:
        point_data = kwargs['data']['point']
        cell_data = kwargs['data']['cell']
    nplots=0
    if point_data: nplots += len(point_data)
    if cell_data: nplots += len(cell_data)
    ax.clf()
    ax.subplot(1, 2, 1)
    if suptitle: plt.gcf().suptitle(suptitle)
    # if title: plt.gcf().set_title(title)
    # aspect = (np.max(x)-np.mean(x))/(np.max(y)-np.mean(y))
    count=0
    if point_data:
        for pdn, pd in point_data.items():
            # print("print", pdn)
            assert x.shape == pd.shape
            ax.plot(x, pd, label=pdn)
        ax.legend()
    ax.subplot(1, 2, 2)
    if cell_data:
        xc =  0.5*(x[:-1] + x[1:])
        for cdn, cd in cell_data.items():
            assert xc.shape == cd.shape
            ax.plot(xc, cd, label=cdn)
        ax.legend()





# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh


#=================================================================#
class FemD0(object):
    def __init__(self, mesh=None):
        if mesh is not None:
            self.setMesh(mesh)
    def setMesh(self, mesh):
        self.mesh = mesh
    def computeErrorL2(self, solex, uh):
        x, y, z = self.mesh.pointsc[:,0], self.mesh.pointsc[:,1], self.mesh.pointsc[:,2]
        e = uh - solex(x, y, z)
        return np.sqrt( np.dot(e, self.mesh.dV*e) ), e


#=================================================================#
if __name__ == '__main__':
    trimesh = SimplexMesh(geomname="backwardfacingstep", hmean=0.3)
    fem = FemD0(trimesh)
    raise NotImplementedError



import numpy as np
from simfempy import fems
from simfempy.applications.application import Application
from simfempy.fems.rt0 import RT0
from simfempy.tools.analyticalfunction import AnalyticalFunction

#=================================================================#
class Transport(Application):
    """
    Class for the (stationary) transport equation
    $$
    alpha u + \div(beta u) = f   domain
    beta\cdot n = g              bdry
    $$
    After initialization, the function setMesh(mesh) has to be called
    Then, solve() solves the stationary problem
    Parameters in the constructor:
        fem: only p1 or cr1
        problemdata
        method
        plotk
    Paramaters used from problemdata:
        alpha : global constant from problemdata.paramglobal
        beta : RT0 field
    Possible parameters for computaion of postprocess:
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        beta_given = self.problemdata.params.fct_glob['beta']
        if not isinstance(beta_given,list):
            p = "problemdata.params.fct_glob['beta']"
            raise ValueError(f"need '{p}' as a list of length dim of str or AnalyticalSolution")
        elif isinstance(beta_given[0],str):
            self.problemdata.params.fct_glob['beta'] = [AnalyticalFunction(expr=e) for e in beta_given]
        if 'linearsolver' in kwargs: self.linearsolver = kwargs.pop('linearsolver')
        else: self.linearsolver = 'umf'
        fem = 'p1'
        if 'fem' in kwargs: fem = kwargs.pop('fem')
        if fem == 'p1':
            self.fem = fems.p1.P1()
        else:
            raise ValueError("unknown fem '{}'".format(fem))
        if 'method' in kwargs:
            self.method = kwargs.pop('method')
        else:
            self.method="supg"
    def _checkProblemData(self):
        pass
    def defineRhsAnalyticalSolution(self, solexact):
        def _fctu(x, y, z):
            alpha = self.problemdata.params.scal_glob['alpha']
            beta = self.problemdata.params.fct_glob['beta']
            rhs = alpha*solexact(x,y,z)
            for i in range(self.mesh.dimension):
                rhs += beta[i](x,y,z) * solexact.d(i, x, y, z)
            return rhs
        return _fctu

    def defineBdryFctAnalyticalSolution(self, color):
        return self.dirichletfct()
    def setMesh(self, mesh):
        super().setMesh(mesh)
        self._checkProblemData()
        self.fem.setMesh(self.mesh)
        betafct = self.problemdata.params.fct_glob['beta']
        rt = RT0(mesh)
        self.beta = rt.interpolate(betafct)
        self.xd, self.lamdbeta, self.delta = self.fem.downWind(self.beta, method=self.method)
        self.betaC = rt.toCell(self.beta)
    def computeMatrix(self):
        A =  self.fem.comptuteMatrixTransport(self.beta, self.betaC, self.lamdbeta)
        # print(f"{self.fem.nunknowns()=} {A.data=}")
        return A
    def computeRhs(self, u=None):
        b = np.zeros(self.fem.nunknowns())
        if 'rhs' in self.problemdata.params.fct_glob:
            fp1 = self.fem.interpolate(self.problemdata.params.fct_glob['rhs'])
            # print(f"{fp1=}")
            # A = self.fem.computeMassMatrixSupg(self.xd)
            # b += A.dot(fp1)
            self.fem.massDot(b, fp1)
            self.fem.massDotSupg(b, fp1, self.xd)
        if self.problemdata.solexact:
            f = self.fem.interpolate(self.problemdata.solexact)
        self.fem.massDotBoundary(b, f, coeff=-np.minimum(self.beta,0))
        # print(f"{b=}")
        return b,u
    def postProcess(self, u):
        point_data, side_data, cell_data, global_data = {}, {}, {}, {}
        point_data['U'] = self.fem.tonode(u)
        if self.problemdata.solexact:
            global_data['error'] = {}
            global_data['error']['pcL2'], ec = self.fem.computeErrorL2Cell(self.problemdata.solexact, u)
            global_data['error']['pnL2'], en = self.fem.computeErrorL2Node(self.problemdata.solexact, u)
            cell_data['E'] = ec
        return point_data, side_data, cell_data, global_data


#=================================================================#
if __name__ == '__main__':
    print("Pas de test")



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np

#=================================================================#
class IterationCounter(object):
    """
    Simple class for information on iterative solver
    """
    def __init__(self, disp=20, name="", verbose=1):
        self.disp = disp
        self.name = name
        self.verbose = verbose
        self.niter = 0
    def __call__(self, rk=None):
        # if self.disp and self.niter%self.disp==0:
        #     print('iter({}) {:4d}\trk = {}'.format(self.name, self.niter, str(rk)))
        self.niter += 1
    def __del__(self):
        if self.verbose: print('niter ({}) {:4d}'.format(self.name, self.niter))



import numpy as np
import scipy.linalg as linalg
import scipy.sparse
import scipy.sparse.linalg as splinalg
from simfempy import solvers
import simfempy.tools.iterationcounter


#=================================================================#
class LaplaceMixed(solvers.solver.Application):
    """
    """
    def defineRhsAnalyticalSolution(self, solexact):
        # constant diffusion
        def _fctu(x, y, z):
            rhs = np.zeros(x.shape[0])
            if self.beta is not None:
                beta = self.beta(x,y,z)
                for i in range(self.mesh.dimension):
                    rhs += beta[i] * solexact.d(i, x, y, z)
            for i in range(self.mesh.dimension):
                rhs -= self.diff(0)*solexact.dd(i, i, x, y, z)
            return rhs
        return _fctu

    def defineNeumannAnalyticalSolution(self, solexact, color):
        def _fctneumann(x, y, z, nx, ny, nz):
            rhs = np.zeros(x.shape[0])
            normals = nx, ny, nz
            for i in range(self.mesh.dimension):
                rhs += self.diff(0)*solexact.d(i, x, y, z) * normals[i]
            return rhs
        return _fctneumann

    def defineRobinAnalyticalSolution(self, solexact, color):
        return solexact

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.linearsolver = "gmres"
        if 'linearsolver' in kwargs:
            self.linearsolver = kwargs.pop('linearsolver')
        if 'massproj' in kwargs: massproj = kwargs.pop('massproj')
        else: massproj=None
        if 'fem' in kwargs:
            fem = kwargs.pop('fem')
            if fem =='bv0':
                self.femv = simfempy.fems.fembv0.FemBV0()
            else:
                self.femv = simfempy.fems.femrt0.FemRT0(massproj=massproj)
        else:
            self.femv = simfempy.fems.femrt0.FemRT0(massproj=massproj)
        if 'diff' in kwargs:
            self.diff = np.vectorize(kwargs.pop('diff'))
        else:
            self.diff = np.vectorize(lambda i: 1)
            # self.diff = np.vectorize(lambda i: 0.123)
        if 'beta' in kwargs:
            self.beta = np.vectorize(kwargs.pop('beta'))
        else:
            self.beta = None
        if 'method' in kwargs:
            self.method = kwargs.pop('method')
        else:
            self.method="trad"
        if 'plotdiff' in kwargs:
            self.plotdiff = kwargs.pop('plotdiff')
        else:
            self.plotdiff = False
        if 'mesh' in kwargs:
            self.setMesh(kwargs.pop('mesh'))

    def setMesh(self, mesh):
        super().setMesh(mesh)
        self.femv.setMesh(mesh)
        self.diffcell = self.diff(self.mesh.cell_labels)
        self.diffcellinv = 1/self.diffcell

    def postProcess(self, u):
        nfaces, dim =  self.mesh.nfaces, self.mesh.dimension
        info = {}
        cell_data = {'p': u[nfaces:]}
        vc = self.femv.toCell(u[:nfaces])
        pn = self.femv.reconstruct(u[nfaces:], vc, self.diffcellinv)
        for i in range(dim):
            cell_data['v{:1d}'.format(i)] = vc[i::dim]
        point_data = {}
        point_data['p_1'] = pn
        if self.problemdata.solexact:
            err, pe, vexx = self.computeError(self.problemdata.solexact, u[nfaces:], vc, pn)
            # print("vexx", vexx)
            # print("u[:nfaces]", u[:nfaces])
            # print("vc[i::dim]", vc[i::dim])
            cell_data['perr'] = np.abs(pe - u[nfaces:])
            for i in range(dim):
                cell_data['verrx{:1d}'.format(i)] = np.abs(vexx[i] - vc[i::dim])
            info['error'] = err
        info['postproc'] = {}
        if self.problemdata.postproc:
            for name, type in self.problemdata.postproc.type.items():
                colors = self.problemdata.postproc.colors(name)
                if type == "bdrymean":
                    info['postproc'][name] = self.computeBdryMean(pn, colors)
                elif type == "bdryfct":
                    info['postproc'][name] = self.computeBdryFct(u, type, colors)
                elif type == "bdrydn":
                    info['postproc'][name] = self.computeBdryDn(u, colors)
                elif type == "pointvalues":
                    info['postproc'][name] = self.computePointValues(u, type, colors)
                else:
                    raise ValueError("unknown postprocess '{}' for key '{}'".format(type, colors))
        if self.plotdiff: cell_data['diff'] = self.diffcell
        return point_data, cell_data, info

    def computeBdryDn(self, u, data):
        colors = [int(x) for x in data.split(',')]
        flux, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            flux[i] = np.sum(dS*u[faces])
        return flux
        return flux/omega

    def computeBdryMean(self, pn, data):
        colors = [int(x) for x in data.split(',')]
        mean, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            mean[i] = np.sum(dS*np.mean(pn[self.mesh.faces[faces]],axis=1))
        return mean/omega

    def computeError(self, solexact, p, vc, pn):
        nfaces, dim =  self.mesh.nfaces, self.mesh.dimension
        errors = {}
        xc, yc, zc = self.mesh.pointsc.T
        pex = solexact(xc, yc, zc)
        errp = np.sqrt(np.sum((pex-p)**2* self.mesh.dV))
        errv = 0
        vexx=[]
        for i in range(dim):
            solxi = self.diffcell*solexact.d(i, xc, yc, zc)
            errv += np.sum( (solxi-vc[i::dim])**2* self.mesh.dV)
            vexx.append(solxi)
        errv = np.sqrt(errv)
        errors['pcL2'] = errp
        errors['vcL2'] = errv

        x, y, z = self.mesh.points.T
        epn = solexact(x, y, z) - pn
        epn = epn**2
        epn= np.mean(epn[self.mesh.simplices], axis=1)
        errors['pnL2'] = np.sqrt(np.sum(epn* self.mesh.dV))
        return errors, pex, vexx

    def computeRhs(self, u=None):
        xf, yf, zf = self.mesh.pointsf.T
        xc, yc, zc = self.mesh.pointsc.T
        normals =  self.mesh.normals
        bsides = np.zeros(self.mesh.nfaces)
        bcells = np.zeros(self.mesh.ncells)
        if self.problemdata.rhs:
            bcells = -self.problemdata.rhs(xc, yc, zc) * self.mesh.dV
        bdrycond = self.problemdata.bdrycond
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] not in ["Dirichlet","Robin"]: continue
            # nx, ny, nz = np.mean(normals[faces], axis=0)
            # ud = bdrycond.fct[color](xf[faces], yf[faces], zf[faces], nx, ny, nz)
            ud = bdrycond.fct[color](xf[faces], yf[faces], zf[faces])
            # bsides[faces] = linalg.norm(self.mesh.normals[faces],axis=1) * ud
            bsides[faces] = self.femv.rhsDirichlet(faces, ud)
            if self.beta is not None:
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                normalsS = normalsS / dS[:, np.newaxis]
                xf, yf, zf = self.mesh.pointsf[faces].T
                beta = np.array(self.beta(xf, yf, zf))
                # print("beta", beta)
                # print("normalsS", normalsS.T)
                bn = np.einsum("ij,ij->j", beta, normalsS.T)
                # print("bn", bn)
                bn[bn<=0] = 0
                cells = self.mesh.cellsOfFaces[faces,0]
                bcells[cells] += bn*ud*dS

        help = np.zeros(self.mesh.nfaces)
        if hasattr(self.problemdata.bdrycond,'fctexact'):
            for color in self.bdrydata.colorsneum:
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                normalsS = normalsS / dS[:, np.newaxis]
                xf, yf, zf = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS.T
                help[faces] += self.problemdata.bdrycond.fctexact["Neumann"](xf, yf, zf, nx, ny, nz)
            bsides[self.bdrydata.facesinner] -= self.bdrydata.A_inner_neum*help[self.bdrydata.facesneumann]
            bsides[self.bdrydata.facesneumann] += self.bdrydata.A_neum_neum*help[self.bdrydata.facesneumann]
            bcells -= self.bdrydata.B_inner_neum*help[self.bdrydata.facesneumann]

        # for robin-exactsolution
        if self.problemdata.bdrycond.hasExactSolution():
            for color, faces in self.mesh.bdrylabels.items():
                if self.problemdata.bdrycond.type[color] != "Robin": continue
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS,axis=1)
                normalsS = normalsS/dS[:,np.newaxis]
                xf, yf, zf = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS.T
                bsides[faces] += self.problemdata.bdrycond.fctexact["Neumann"](xf, yf, zf, nx, ny, nz) * dS/self.problemdata.bdrycond.param[color]
        b = np.concatenate((bsides, bcells))
        if u is None: u = np.zeros_like(b)
        else: assert u.shape == b.shape
        return b,u

    def matrix(self):
        A = self.femv.constructMass(self.diffcellinv)
        A += self.femv.constructRobin(self.problemdata.bdrycond, "Robin")
        B = self.femv.constructDiv()
        self.bdrydata, A,B = self.femv.matrixNeumann(A, B, self.problemdata.bdrycond)
        return A,B

    def _to_single_matrix(self, Ain):
        A, B = Ain
        ncells = self.mesh.ncells
        help = np.zeros((ncells))
        help = scipy.sparse.dia_matrix((help, 0), shape=(ncells, ncells))
        A1 = scipy.sparse.hstack([A, B.T])
        A2 = scipy.sparse.hstack([B, help])
        Aall = scipy.sparse.vstack([A1, A2])
        return Aall.tocsr()

    def linearSolver(self, Ain, bin, u=None, solver = None, verbose=0):
        if solver is None: solver = self.linearsolver
        if solver == 'umf':
            # print("bin", bin)
            Aall = self._to_single_matrix(Ain)
            u =  splinalg.spsolve(Aall, bin, permc_spec='COLAMD'), 1
            # print("u", u)
            return u
        elif solver == 'gmres':
            nfaces, ncells = self.mesh.nfaces, self.mesh.ncells
            counter = simfempy.tools.iterationcounter.IterationCounter(name=solver, verbose=verbose)
            # Aall = self._to_single_matrix(Ain)
            # M2 = splinalg.spilu(Aall, drop_tol=0.2, fill_factor=2)
            # M_x = lambda x: M2.solve(x)
            # M = splinalg.LinearOperator(Aall.shape, M_x)
            A, B = Ain
            A, B = A.tocsr(), B.tocsr()
            D = scipy.sparse.diags(1/A.diagonal(), offsets=(0), shape=(nfaces,nfaces))
            S = -B*D*B.T
            # print("S", S)


            # Dex = scipy.sparse.linalg.inv(A)
            # print("Sex", -B*Dex*B.T)


            import pyamg
            config = pyamg.solver_configuration(S, verb=False)
            ml = pyamg.rootnode_solver(S, B=config['B'], smooth='energy')
            # Ailu = splinalg.spilu(A, drop_tol=0.2, fill_factor=2)
            def amult(x):
                v,p = x[:nfaces],x[nfaces:]
                return np.hstack( [A.dot(v) + B.T.dot(p), B.dot(v)])
            Amult = splinalg.LinearOperator(shape=(nfaces+ncells,nfaces+ncells), matvec=amult)
            def pmult(x):
                v,p = x[:nfaces],x[nfaces:]
                w = D.dot(v)
                # w = Ailu.solve(v)
                q = ml.solve(p - B.dot(w), maxiter=1, tol=1e-16)
                w = w - D.dot(B.T.dot(q))
                # w = w - Ailu.solve(B.T.dot(q))
                return np.hstack( [w, q] )
            P = splinalg.LinearOperator(shape=(nfaces+ncells,nfaces+ncells), matvec=pmult)
            u,info = splinalg.lgmres(Amult, bin, M=P, callback=counter, atol=1e-12, tol=1e-12, inner_m=10, outer_k=4)
            if info: raise ValueError("no convergence info={}".format(info))
            # print("u", u)
            return u, counter.niter
        else:
            raise NotImplementedError("solver '{}' ".format(solver))

#=================================================================#
if __name__ == '__main__':
    import simfempy
    from simfempy.meshes.simplexmesh import SimplexMesh
    from simfempy.meshes import plotmesh
    import pygmsh
    import matplotlib.pyplot as plt
    geometry = pygmsh.built_in.Geometry()
    a = 1.0
    h = 2
    p = geometry.add_rectangle(xmin=-a, xmax=a, ymin=-a, ymax=a, z=0, lcar=h)
    geometry.add_physical(p.surface, label=100)
    for i in range(4): geometry.add_physical(p.line_loop.lines[i], label=1000 + i)
    mesh = pygmsh.generate_mesh(geometry, verbose=False)
    mesh = SimplexMesh(mesh=mesh, hmean=h)

    data = simfempy.applications.problemdata.ProblemData()
    bdrycond =  data.bdrycond
    postproc = data.postproc
    bdrycond.set("Dirichlet", [1000, 1001, 1002, 1003])
    data.params.scal_glob['kheat'] = 0.123
    laplace = LaplaceMixed(mesh=mesh, problemdata=data, showmesh=False)
    exactsolution = "Linear"
    problemdata = laplace.generatePoblemDataForAnalyticalSolution(exactsolution=exactsolution, problemdata=data, random=False)
    laplace = LaplaceMixed(problemdata=problemdata, mesh=mesh)
    # print("A=",laplace.A)
    # print("B=",laplace.B)
    point_data, cell_data, info = laplace.solve(dirname="Results")
    print("point_data", point_data)
    print("cell_data", cell_data)
    plotmesh.meshWithData(mesh, point_data=point_data, cell_data=cell_data, title="Test", suptitle="LaplaceMixed")
    plt.show()


import numpy as np
import scipy.sparse
import scipy.linalg as linalg
import scipy.sparse.linalg as splinalg
from simfempy import solvers
from simfempy import fems
import simfempy.tools.analyticalfunction
import simfempy.tools.iterationcounter

# =================================================================#
class Stokes(solvers.solver.Application):
    """
    """
    def generatePoblemData(self, **kwargs):
        self.ncomp = self.mesh.dimension
        return super().generatePoblemData(**kwargs)

    def defineAnalyticalSolution(self, exactsolution, random=True):
        solexact = super().defineAnalyticalSolution(exactsolution=exactsolution, random=random)
        if exactsolution == 'Linear':
            # solexact.append(simfempy.tools.analyticalsolution.AnalyticalFunction('x+y'))
            solexact.append(simfempy.tools.analyticalfunction.AnalyticalSolution3d('0'))
        elif exactsolution == 'Quadratic':
            # solexact[0] = simfempy.tools.analyticalsolution.AnalyticalFunction('x*x-2*y*x')
            # solexact[1] = simfempy.tools.analyticalsolution.AnalyticalFunction('-2*x*y+y*y')
            solexact.append(simfempy.tools.analyticalfunction.AnalyticalSolution3d('x+y'))
        else:
            raise NotImplementedError("unknown function '{}'".format(exactsolution))
        return solexact

    def defineRhsAnalyticalSolution(self, solexact):
        print("self.mu", self.mu(0))
        def _fctv(x, y, z):
            rhs = np.zeros(shape=(self.ncomp, x.shape[0]))
            for i in range(self.ncomp):
                for j in range(self.ncomp):
                    rhs[i] -= self.mu(0)* solexact[i].dd(j, j, x, y, z)
                rhs[i] += solexact[self.ncomp].d(i, x, y, z)
            return rhs
        def _fctp(x, y, z):
            rhs = np.zeros(x.shape[0])
            for i in range(self.ncomp):
                rhs += solexact[i].d(i, x, y, z)
            return rhs
        return _fctv, _fctp

    def defineNeumannAnalyticalSolution(self, solexact):
        def _fctneumann(x, y, z, nx, ny, nz):
            rhs = np.zeros(shape=(self.ncomp, x.shape[0]))
            normals = nx, ny, nz
            for i in range(self.ncomp):
                for j in range(self.ncomp):
                    rhs[i] += self.mu(0) * solexact[i].d(j, x, y, z) * normals[j]
                rhs[i] -= solexact[self.ncomp](x, y, z) * normals[i]
            return rhs
        return _fctneumann

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # if 'geometry' in kwargs:
        #     return
        self.linearsolver = 'gmres'
        if 'fem' in kwargs:
            fem = kwargs.pop('fem')
        else:
            fem = 'cr1'
        if fem == 'p1':
            self.femv = fems.femp1sys.FemP1()
            raise NotImplementedError("p1 not ready")
        elif fem == 'cr1':
            self.femp = fems.femd0.FemD0()
            self.femv = fems.femcr1sys.FemCR1()
        else:
            raise ValueError("unknown fem '{}'".format(fem))
        if hasattr(self,'problemdata') and hasattr(self.problemdata,'mu'):
            self.mu = np.vectorize(self.problemdata.mu)
        else:
            self.mu = np.vectorize(lambda x: 1.0)
        if 'method' in kwargs:
            self.method = kwargs.pop('method')
        else:
            self.method = "trad"
        if 'rhsmethod' in kwargs:
            self.problemdata.rhsmethod = kwargs.pop('rhsmethod')
            if self.problemdata.rhsmethod == "rt":
                self.femrt = fems.femrt0.FemRT0()

    def setMesh(self, mesh):
        super().setMesh(mesh)
        self.femv.setMesh(self.mesh, self.ncomp)
        self.femp.setMesh(self.mesh)
        self.pmean = self.problemdata.bdrycond.type.values() == len(self.problemdata.bdrycond.type)*"Dirichlet"
        self.bdrydata = self.femv.prepareBoundary(self.problemdata.bdrycond, self.problemdata.postproc)
        self.mucell = self.mu(self.mesh.cell_labels)
        self.pstart = self.ncomp*self.mesh.nfaces
        if self.problemdata.rhsmethod == "rt":
            self.femrt.setMesh(self.mesh)
            self.massrt = self.femrt.constructMass()

    def solve(self, iter, dirname):
        return self.solveLinear()

    def computeRhs(self, u=None):
        ncomp, nfaces, ncells = self.ncomp, self.mesh.nfaces, self.mesh.ncells
        if self.pmean: nall = nfaces * ncomp + self.mesh.ncells +1
        else: nall = nfaces * ncomp + self.mesh.ncells
        b = np.zeros(nall)
        rhsv, rhsp = self.problemdata.rhs
        xf, yf, zf = self.mesh.pointsf.T
        xc, yc, zc = self.mesh.pointsc.T
        if rhsp:
            b[self.pstart:self.pstart + ncells] = self.mesh.dV * np.array(rhsp(xc, yc, zc))
        rhsall = np.array(rhsv(xf, yf, zf))
        if self.problemdata.rhsmethod=='rt':
            normals, sigma, dV = self.mesh.normals, self.mesh.sigma, self.mesh.dV
            dS = linalg.norm(normals, axis=1)
            rhsn = np.zeros(nfaces)
            for i in range(ncomp):
                rhsn += rhsall[i] * normals[:,i] / dS
            rhsn = self.massrt*rhsn
            for i in range(ncomp):
                b[i * nfaces:(i + 1) * nfaces] = rhsn*normals[:,i]/dS
        elif self.problemdata.rhsmethod=='cr':
            for i in range(ncomp):
                b[i * nfaces:(i + 1) * nfaces] = self.femv.massmatrix * rhsall[i]
        else:
            raise ValueError("don't know rhsmethod='{}'".format(self.problemdata.rhsmethod))
        normals = self.mesh.normals
        for color, faces in self.mesh.bdrylabels.items():
            condition = self.problemdata.bdrycond.type[color]
            if condition == "Neumann":
                normalsS = normals[faces]
                dS = linalg.norm(normalsS,axis=1)
                xf, yf, zf = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS[:,0]/dS, normalsS[:,1]/dS, normalsS[:,2]/dS
                if not color in self.problemdata.bdrycond.fct.keys(): continue
                neumanns = self.problemdata.bdrycond.fct[color](xf, yf, zf, nx, ny, nz)
                for i in range(ncomp):
                    bS = dS * neumanns[i]
                    indices = i*nfaces + faces
                    b[indices] += bS
        return self.vectorDirichlet(b, u)

    def matrix(self):
        nfaces, facesOfCells = self.mesh.nfaces, self.mesh.facesOfCells
        nnodes, ncells, ncomp, dV = self.mesh.nnodes, self.mesh.ncells, self.ncomp, self.mesh.dV
        nloc, cellgrads = self.femv.nloc, self.femv.cellgrads
        matA = np.zeros(ncells*nloc*nloc).reshape(ncells, nloc, nloc)
        for i in range(ncomp):
            matA += np.einsum('nk,nl->nkl', cellgrads[:, :, i], cellgrads[:, :, i])
        matA = ( matA.T*dV*self.mucell).T.ravel()
        cols = np.tile(facesOfCells, nloc).reshape(ncells, nloc, nloc)
        rows = cols.swapaxes(1, 2)
        A = scipy.sparse.coo_matrix((matA, (rows.reshape(-1), cols.reshape(-1))), shape=(nfaces, nfaces)).tocsr()
        rowsB = np.repeat(np.arange(ncells), ncomp*nloc).reshape(ncells * nloc, ncomp)
        colsB = np.repeat( facesOfCells, ncomp).reshape(ncells * nloc, ncomp) + nfaces*np.arange(ncomp)
        matB = (cellgrads[:,:,:ncomp].T*dV).T
        B = scipy.sparse.coo_matrix((matB.reshape(-1), (rowsB.reshape(-1), colsB.reshape(-1))), shape=(ncells, nfaces*ncomp)).tocsr()
        (A,B) = self.matrixDirichlet((A,B))
        if self.pmean:
            rows = np.zeros(ncells, dtype=int)
            cols = np.arange(0, ncells)
            C = scipy.sparse.coo_matrix((dV, (rows, cols)), shape=(1, ncells)).tocsr()
            return (A,B,C)
        return (A,B)

    def vectorDirichlet(self, b, u):
        if u is None: u = np.zeros_like(b)
        else: assert u.shape == b.shape
        xf, yf, zf = self.mesh.pointsf.T
        facesdirall, facesinner, colorsdir, facesdirflux = self.bdrydata.facesdirall, self.bdrydata.facesinner, self.bdrydata.colorsdir, self.bdrydata.facesdirflux
        nfaces, ncells, ncomp  = self.mesh.nfaces, self.mesh.ncells, self.femv.ncomp
        self.bdrydata.bsaved = []
        for icomp in range(ncomp):
            self.bdrydata.bsaved.append({})
            for key, faces in facesdirflux.items():
                self.bdrydata.bsaved[icomp][key] = b[icomp*nfaces + faces]
        if self.method == 'trad':
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = self.problemdata.bdrycond.fct[color]
                dirs = dirichlet(xf[faces], yf[faces], zf[faces])
                # print("dirs", dirs)
                for icomp in range(ncomp):
                    b[icomp*nfaces + faces] = dirs[icomp]
                    u[icomp*nfaces + faces] = b[icomp*nfaces + faces]
            for icomp in range(ncomp):
                indin = icomp*nfaces + facesinner
                inddir = icomp*nfaces + facesdirall
                b[indin] -= self.bdrydata.A_inner_dir * b[inddir]
                b[self.pstart:self.pstart+ncells] -= self.bdrydata.B_inner_dir[icomp] * b[inddir]
        else:
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = self.problemdata.bdrycond.fct[color]
                dirs = dirichlet(xf[faces], yf[faces], zf[faces])
                for icomp in range(ncomp):
                    u[icomp*nfaces + faces] = dirs[icomp]
                    b[icomp*nfaces + faces] = 0
            for icomp in range(ncomp):
                indin = icomp*nfaces + facesinner
                inddir = icomp*nfaces + facesdirall
                b[indin] -= self.bdrydata.A_inner_dir * u[inddir]
                b[inddir] += self.bdrydata.A_dir_dir * u[inddir]
        return b, u

    def matrixDirichlet(self, Aall):
        A, B = Aall
        facesdirall, facesinner, colorsdir, facesdirflux = self.bdrydata.facesdirall, self.bdrydata.facesinner, self.bdrydata.colorsdir, self.bdrydata.facesdirflux
        nfaces, ncells, ncomp  = self.mesh.nfaces, self.mesh.ncells, self.femv.ncomp
        self.bdrydata.Asaved = {}
        self.bdrydata.Bsaved = {}
        for key, faces in facesdirflux.items():
            nb = faces.shape[0]
            help = scipy.sparse.dok_matrix((nb, nfaces))
            for i in range(nb): help[i, faces[i]] = 1
            self.bdrydata.Asaved[key] = help.dot(A)
            helpB = scipy.sparse.dok_matrix((ncomp*nfaces, ncomp*nb))
            for icomp in range(ncomp):
                for i in range(nb): helpB[icomp*nfaces + faces[i], icomp*nb + i] = 1
            self.bdrydata.Bsaved[key] = B.dot(helpB)
        self.bdrydata.A_inner_dir = A[facesinner, :][:, facesdirall]
        self.bdrydata.B_inner_dir = []
        for icomp in range(ncomp):
            inddir = icomp * nfaces + facesdirall
            self.bdrydata.B_inner_dir.append(B[:,:][:,inddir])
        if self.method == 'trad':
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = scipy.sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help))
            help = np.zeros((nfaces))
            help[facesdirall] = 1.0
            help = scipy.sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A += help
        else:
            self.bdrydata.A_dir_dir = A[facesdirall, :][:, facesdirall]
            for icomp in range(ncomp):
                indin = icomp*nfaces + facesinner
                inddir = icomp*nfaces + facesdirall
                b[indin] -= self.A_inner_dir * u[inddir]
                b[inddir] += self.A_dir_dir * u[inddir]
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            help2 = np.zeros((nfaces))
            help2[facesdirall] = 1
            help2 = sparse.dia_matrix((help2, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        # B
        help = np.ones((ncomp * nfaces))
        for icomp in range(ncomp):
            help[icomp*nfaces + facesdirall] = 0
        help = scipy.sparse.dia_matrix((help, 0), shape=(ncomp * nfaces, ncomp * nfaces))
        B = B.dot(help)
        return (A,B)

    def boundary(self, Aall, b, u):
        if self.pmean:
            A, B, C = Aall
        else:
            A, B = Aall
        # print("A", A.todense())
        # print("B", B.todense())
        xf, yf, zf = self.mesh.pointsf.T
        facesdirall, facesinner, colorsdir, facesdirflux = self.bdrydata
        nfaces, ncells, ncomp  = self.mesh.nfaces, self.mesh.ncells, self.femv.ncomp
        self.bsaved = []
        self.Asaved = {}
        self.Bsaved = {}
        for key, faces in facesdirflux.items():
            nb = faces.shape[0]
            help = scipy.sparse.dok_matrix((nb, nfaces))
            for i in range(nb): help[i, faces[i]] = 1
            self.Asaved[key] = help.dot(A)
            helpB = scipy.sparse.dok_matrix((ncomp*nfaces, ncomp*nb))
            for icomp in range(ncomp):
                for i in range(nb): helpB[icomp*nfaces + faces[i], icomp*nb + i] = 1
            self.Bsaved[key] = B.dot(helpB)
        self.A_inner_dir = A[facesinner, :][:, facesdirall]
        for icomp in range(ncomp):
            self.bsaved.append({})
            for key, faces in facesdirflux.items():
                self.bsaved[icomp][key] = b[icomp*nfaces + faces]
        if self.method == 'trad':
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = self.problemdata.bdrycond.fct[color]
                dirs = dirichlet(xf[faces], yf[faces], zf[faces])
                # print("dirs", dirs)
                for icomp in range(ncomp):
                    b[icomp*nfaces + faces] = dirs[icomp]
                    u[icomp*nfaces + faces] = b[icomp*nfaces + faces]
            for icomp in range(ncomp):
                indin = icomp*nfaces + facesinner
                inddir = icomp*nfaces + facesdirall
                b[indin] -= self.A_inner_dir * b[inddir]
                b[self.pstart:self.pstart+ncells] -= B[:,:][:,inddir] * b[inddir]
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = scipy.sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help))
            help = np.zeros((nfaces))
            help[facesdirall] = 1.0
            help = scipy.sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A += help
        else:
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = self.problemdata.bdrycond.fct[color]
                dirs = dirichlet(xf[faces], yf[faces], zf[faces])
                for icomp in range(ncomp):
                    u[icomp*nfaces + faces] = dirs[icomp]
                    b[icomp*nfaces + faces] = 0
            self.A_dir_dir = A[facesdirall, :][:, facesdirall]
            for icomp in range(ncomp):
                indin = icomp*nfaces + facesinner
                inddir = icomp*nfaces + facesdirall
                b[indin] -= self.A_inner_dir * u[inddir]
                b[inddir] += self.A_dir_dir * u[inddir]
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            help2 = np.zeros((nfaces))
            help2[facesdirall] = 1
            help2 = sparse.dia_matrix((help2, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        # B
        help = np.ones((ncomp * nfaces))
        for icomp in range(ncomp):
            help[icomp*nfaces + facesdirall] = 0
        help = scipy.sparse.dia_matrix((help, 0), shape=(ncomp * nfaces, ncomp * nfaces))
        B = B.dot(help)
        # print("B", B[0])
        # print("after A", A.todense())
        # print("after B", B.todense())
        if self.pmean:
            return (A, B, C), b, u
        return (A,B), b, u

    def computeBdryMean(self, u, data):
        nfaces, ncomp  = self.mesh.nfaces, self.femv.ncomp
        colors = [int(x) for x in data.split(',')]
        mean, omega = np.zeros(shape=(ncomp,len(colors))), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            for icomp in range(ncomp):
                mean[icomp,i] += np.sum(dS * u[icomp*nfaces + faces])
        return mean/omega
        # colors = [int(x) for x in data.split(',')]
        # mean, omega = [0 for i in range(self.ncomp)], 0
        # for color in colors:
        #     faces = self.mesh.bdrylabels[color]
        #     normalsS = self.mesh.normals[faces]
        #     dS = linalg.norm(normalsS, axis=1)
        #     for i in range(ncomp):
        #         mean[i] += np.sum(dS * u[i*nfaces + faces])
        # return mean

    def computeBdryDn(self, u, data, bdrydata, bdrycond):
        nfaces, ncells, ncomp, pstart  = self.mesh.nfaces, self.mesh.ncells, self.femv.ncomp, self.pstart
        colors = [int(x) for x in data.split(',')]
        flux, omega = np.zeros(shape=(ncomp,len(colors))), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            if bdrycond.type[color] == "Dirichlet":
                As = bdrydata.Asaved[color]
                Bs = bdrydata.Bsaved[color]
                pcontrib = Bs.T * u[pstart: pstart + ncells]
                nb = Bs.shape[1] // ncomp
                for icomp in range(ncomp):
                    bs = bdrydata.bsaved[icomp][color]
                    res = bs - As * u[icomp * nfaces:(icomp + 1) * nfaces]
                    flux[icomp, i] = np.sum(res+pcontrib[icomp*nb:(icomp+1)*nb])
            else:
                raise NotImplementedError("computeBdryDn for condition '{}'".format(bdrycond.type[color]))
        return flux

        # flux = []
        # pcontrib = self.bdrydata.Bsaved[key].T*u[pstart: pstart+ncells]
        # nb = self.bdrydata.Bsaved[key].shape[1]//ncomp
        # for icomp in range(ncomp):
        #     res = self.bdrydata.bsaved[icomp][key] - self.bdrydata.Asaved[key] * u[icomp*nfaces:(icomp+1)*nfaces]
        #     flux.append(np.sum(res+pcontrib[icomp*nb:(icomp+1)*nb]))
        # return flux

    def computeErrorL2V(self, solex, uh):
        nfaces, ncomp  = self.mesh.nfaces, self.femv.ncomp
        xf, yf, zf = self.mesh.pointsf.T
        e = []
        err = []
        for icomp in range(self.ncomp):
            e.append(solex[icomp](xf, yf, zf) - uh[icomp*nfaces:(icomp+1)*nfaces])
            err.append(np.sqrt(np.dot(e[icomp], self.femv.massmatrix * e[icomp])))
        return err, e

    def postProcess(self, u):
        nfaces, ncomp  = self.mesh.nfaces, self.femv.ncomp
        info = {}
        cell_data = {}
        point_data = {}
        for icomp in range(ncomp):
            point_data['V_{:02d}'.format(icomp)] = self.femv.tonode(u[icomp*nfaces:(icomp+1)*nfaces])
        p = u[self.pstart:self.pstart+self.mesh.ncells]
        cell_data['P'] = p
        if self.problemdata.solexact:
            info['error'] = {}
            errv, ev = self.computeErrorL2V(self.problemdata.solexact, u[:self.pstart])
            errp, ep = self.femp.computeErrorL2(self.problemdata.solexact[-1], p)
            info['error']['L2-V'] = np.sum(errv)
            info['error']['L2-P'] = np.sum(errp)
            for icomp in range(self.ncomp):
                point_data['E_V{:02d}'.format(icomp)] = self.femv.tonode(ev[icomp])
            cell_data['E_P'] = ep
        info['postproc'] = {}
        for key, val in self.problemdata.postproc.items():
            type, data = val.split(":")
            if type == "bdrymean":
                mean = self.computeBdryMean(u, data)
                assert len(mean) == self.ncomp
                for icomp in range(self.ncomp):
                    info['postproc']["{}_{:02d}".format(key, icomp)] = mean[icomp]
            elif type == "bdrydn":
                flux = self.computeBdryDn(u, data, self.bdrydata, self.problemdata.bdrycond)
                assert len(flux) == self.ncomp
                for icomp in range(self.ncomp):
                    info['postproc']["{}_{:02d}".format(key, icomp)] = flux[icomp]
            else:
                raise ValueError("unknown postprocess {}".format(key))
        return point_data, cell_data, info

    def _to_single_matrix(self, Ain):
        import scipy.sparse
        # print("Ain", Ain)
        if self.pmean:
            A, B, C = Ain
        else:
            A, B = Ain
        ncells, nfaces, = self.mesh.ncells, self.mesh.nfaces
        nullV = scipy.sparse.dia_matrix((np.zeros(nfaces), 0), shape=(nfaces, nfaces))
        nullP = scipy.sparse.dia_matrix((np.zeros(ncells), 0), shape=(ncells, ncells))
        if self.ncomp==2:
            A1 = scipy.sparse.hstack([A, nullV])
            A2 = scipy.sparse.hstack([nullV, A])
            AV = scipy.sparse.vstack([A1, A2])
        else:
            A1 = scipy.sparse.hstack([A, nullV, nullV])
            A2 = scipy.sparse.hstack([nullV, A, nullV])
            A3 = scipy.sparse.hstack([nullV, nullV, A])
            AV = scipy.sparse.vstack([A1, A2, A3])
        A1 = scipy.sparse.hstack([AV, -B.T])
        A2 = scipy.sparse.hstack([B, nullP])
        Aall = scipy.sparse.vstack([A1, A2])

        if self.pmean:
            rows = np.zeros(nfaces, dtype=int)
            cols = np.arange(0, nfaces)
            nullV = sparse.coo_matrix((np.zeros(nfaces), (rows, cols)), shape=(1, nfaces)).tocsr()
            if self.ncomp == 2:
                CL = scipy.sparse.hstack([nullV, nullV, C])
            else:
                CL = scipy.sparse.hstack([nullV, nullV, nullV, C])
            Abig = scipy.sparse.hstack([Aall,CL.T])
            nullL = scipy.sparse.dia_matrix((np.zeros(1), 0), shape=(1, 1))
            Cbig = scipy.sparse.hstack([CL,nullL])
            Aall = scipy.sparse.vstack([Abig, Cbig])

        return Aall.tocsr()

    def linearSolver(self, Ain, bin, u=None, solver='umf'):
        if solver == 'umf':
            Aall = self._to_single_matrix(Ain)
            u =  splinalg.spsolve(Aall, bin, permc_spec='COLAMD')
            return u, 1
        elif solver == 'gmres':
            nfaces, ncells, ncomp, pstart = self.mesh.nfaces, self.mesh.ncells, self.ncomp, self.pstart
            counter = simfempy.tools.iterationcounter.IterationCounter(name=solver)
            if self.pmean:
                A, B, C = Ain
                nall = ncomp*nfaces + ncells + 1
                BP = scipy.sparse.diags(1/self.mesh.dV, offsets=(0), shape=(ncells, ncells))
                CP = splinalg.inv(C*BP*C.T)
                import pyamg
                config = pyamg.solver_configuration(A, verb=False)
                API = pyamg.rootnode_solver(A, B=config['B'], smooth='energy')
                def amult(x):
                    v, p, lam = x[:pstart], x[pstart:pstart+ncells], x[pstart+ncells:]
                    w = -B.T.dot(p)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] += A.dot(v[i*nfaces: (i+1)*nfaces])
                    q = B.dot(v)+C.T.dot(lam).ravel()
                    return np.hstack([w, q, C.dot(p)])
                Amult = splinalg.LinearOperator(shape=(nall, nall), matvec=amult)
                def pmult(x):
                    v, p, lam = x[:pstart], x[pstart:pstart+ncells], x[pstart+ncells:]
                    w = np.zeros_like(v)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] = API.solve(v[i*nfaces: (i+1)*nfaces], maxiter=1, tol=1e-16)
                    q = BP.dot(p-B.dot(w))
                    mu = CP.dot(lam-C.dot(q)).ravel()
                    q += BP.dot(C.T.dot(mu).ravel())
                    h = B.T.dot(q)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] += API.solve(h[i*nfaces: (i+1)*nfaces], maxiter=1, tol=1e-16)
                    return np.hstack([w, q, mu])
                P = splinalg.LinearOperator(shape=(nall, nall), matvec=pmult)
                u, info = splinalg.lgmres(Amult, bin, M=P, callback=counter, atol=1e-14, tol=1e-14, inner_m=10, outer_k=4)
                if info: raise ValueError("no convergence info={}".format(info))
                return u, counter.niter
            else:
                A, B = Ain
                nall = ncomp*nfaces + ncells
                BP = scipy.sparse.diags(1/self.mesh.dV, offsets=(0), shape=(ncells, ncells))
                import pyamg
                config = pyamg.solver_configuration(A, verb=False)
                API = pyamg.rootnode_solver(A, B=config['B'], smooth='energy')
                def amult(x):
                    v, p= x[:pstart], x[pstart:pstart+ncells]
                    w = -B.T.dot(p)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] += A.dot(v[i*nfaces: (i+1)*nfaces])
                    q = B.dot(v)
                    return np.hstack([w, q])
                Amult = splinalg.LinearOperator(shape=(nall, nall), matvec=amult)
                def pmult(x):
                    v, p = x[:pstart], x[pstart:pstart+ncells]
                    w = np.zeros_like(v)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] = API.solve(v[i*nfaces: (i+1)*nfaces], maxiter=1, tol=1e-16)
                    q = BP.dot(p-B.dot(w))
                    h = B.T.dot(q)
                    for i in range(ncomp):
                        w[i*nfaces: (i+1)*nfaces] += API.solve(h[i*nfaces: (i+1)*nfaces], maxiter=1, tol=1e-16)
                    return np.hstack([w, q])
                P = splinalg.LinearOperator(shape=(nall, nall), matvec=pmult)
                u, info = splinalg.lgmres(Amult, bin, M=P, callback=counter, atol=1e-14, tol=1e-14, inner_m=10, outer_k=4)
                if info: raise ValueError("no convergence info={}".format(info))
                return u, counter.niter
        else:
            raise ValueError("unknown solve '{}'".format(solver))



import numpy as np
from simfempy import fems
from simfempy.applications.application import Application
from simfempy.fems.rt0 import RT0
from simfempy.tools.analyticalfunction import AnalyticalFunction

#=================================================================#
class Heat(Application):
    """
    Class for the (stationary) heat equation
    $$
    rhoCp (T_t + v\cdot\nabla T) -\div(kheat \nabla T) = f         domain
    kheat\nabla\cdot n + alpha T = g  bdry
    $$
    After initialization, the function setMesh(mesh) has to be called
    Then, solve() solves the stationary problem
    Parameters in the constructor:
        fem: only p1 or cr1
        problemdata
        method
        masslumpedbdry, masslumpedvol
    Paramaters used from problemdata:
        rhocp
        kheat
        reaction
        alpha
        they can either be given as global constant, cell-wise constants, or global function
        - global constant is taken from problemdata.paramglobal
        - cell-wise constants are taken from problemdata.paramcells
        - problemdata.paramglobal is taken from problemdata.datafct and are called with arguments (color, xc, yc, zc)
    Possible parameters for computaion of postprocess:
        errors
        bdry_mean: computes mean temperature over boundary parts according to given color
        bdry_nflux: computes mean normal flux over boundary parts according to given color
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.linearsolver = kwargs.pop('linearsolver', 'pyamg')
        self.method = kwargs.pop('method', "trad")
        self.masslumpedbdry = kwargs.pop('masslumpedbdry', True)
        self.masslumpedvol = kwargs.pop('masslumpedvol', True)

        fem = kwargs.pop('fem','p1')
        if fem == 'p1': self.fem = fems.p1.P1()
        elif fem == 'cr1': self.fem = fems.femcr1.FemCR1()
        else: raise ValueError("unknown fem '{}'".format(fem))
        if 'convection' in self.problemdata.params.fct_glob.keys():
            convection_given = self.problemdata.params.fct_glob['convection']
            if not isinstance(convection_given, list):
                p = "problemdata.params.fct_glob['convection']"
                raise ValueError(f"need '{p}' as a list of length dim of str or AnalyticalSolution")
            elif isinstance(convection_given[0],str):
                self.problemdata.params.fct_glob['convection'] = [AnalyticalFunction(expr=e) for e in convection_given]
            else:
                if not isinstance(convection_given[0], AnalyticalFunction):
                    raise ValueError(f"convection should be given as 'str' and not '{type(convection_given[0])}'")
    def _checkProblemData(self):
        self.problemdata.check(self.mesh)
        bdrycond = self.problemdata.bdrycond
        for color in self.mesh.bdrylabels:
            if not color in bdrycond.type: raise ValueError(f"color={color} not in bdrycond={bdrycond}")
            if bdrycond.type[color] in ["Robin"]:
                if not color in bdrycond.param:
                    raise ValueError(f"Robin condition needs paral 'alpha' color={color} bdrycond={bdrycond}")
    def defineRhsAnalyticalSolution(self, solexact):
        def _fctu(x, y, z):
            kheat = self.problemdata.params.scal_glob['kheat']
            rhs = np.zeros(x.shape[0])
            for i in range(self.mesh.dimension):
                rhs -= kheat * solexact.dd(i, i, x, y, z)
            return rhs
        return _fctu
    def defineNeumannAnalyticalSolution(self, problemdata, color):
        solexact = problemdata.solexact
        def _fctneumann(x, y, z, nx, ny, nz):
            kheat = self.problemdata.params.scal_glob['kheat']
            rhs = np.zeros(x.shape[0])
            normals = nx, ny, nz
            for i in range(self.mesh.dimension):
                rhs += kheat * solexact.d(i, x, y, z) * normals[i]
            return rhs
        return _fctneumann
    def defineRobinAnalyticalSolution(self, problemdata, color):
        solexact = problemdata.solexact
        alpha = problemdata.bdrycond.param[color]
        # alpha = 1
        def _fctrobin(x, y, z, nx, ny, nz):
            kheat = self.problemdata.params.scal_glob['kheat']
            rhs = np.zeros(x.shape[0])
            normals = nx, ny, nz
            # print(f"{alpha=}")
            # rhs += alpha*solexact(x, y, z)
            rhs += solexact(x, y, z)
            for i in range(self.mesh.dimension):
                # rhs += kheat * solexact.d(i, x, y, z) * normals[i]
                rhs += kheat * solexact.d(i, x, y, z) * normals[i]/alpha
            return rhs
        return _fctrobin
    def setParameter(self, paramname, param):
        if paramname == "dirichlet_al": self.fem.dirichlet_al = param
        else:
            if not hasattr(self, self.paramname):
                raise NotImplementedError("{} has no paramater '{}'".format(self, self.paramname))
            cmd = "self.{} = {}".format(self.paramname, param)
            eval(cmd)
    def solve(self, iter, dirname): return self.static(iter, dirname)
    def setMesh(self, mesh):
        super().setMesh(mesh)
        # if mesh is not None: self.mesh = mesh
        self._checkProblemData()
        self.fem.setMesh(self.mesh)
        dircols = self.problemdata.bdrycond.colorsOfType("Dirichlet")
        colorsflux = self.problemdata.postproc.colorsOfType("bdry_nflux")
        self.bdrydata = self.fem.prepareBoundary(dircols, colorsflux)
        params = self.problemdata.params
        self.kheatcell = self.compute_cell_vector_from_params('kheat', params)
        self.problemdata.params.scal_glob.setdefault('rhocp',1)
        # TODO: non-constant rhocp
        rhocp = self.problemdata.params.scal_glob.setdefault('rhocp', 1)
        # if not params.paramdefined('rhocp'): params.scal_glob['rhocp'] = 1
        # self.rhocpcell = self.compute_cell_vector_from_params('rhocp', params)
        if 'convection' in self.problemdata.params.fct_glob.keys():
            convectionfct = self.problemdata.params.fct_glob['convection']
            rt = RT0(mesh)
            self.convection = rt.interpolate(convectionfct)
            self.convection *= rhocp
            self.xd, self.lamdconvection, self.delta = self.fem.downWind(self.convection, method="supg")
            self.convectionC = rt.toCell(self.convection)
            conv = self.problemdata.params.fct_glob['convection']
            if len(conv) != self.mesh.dimension: raise ValueError(f"convection has wrong length")
            # for i, c in enumerate(conv):
            #     print(f"{c.fct_x[i]=}")
    def computeMatrix(self):
        bdrycond, method, bdrydata = self.problemdata.bdrycond, self.method, self.bdrydata
        A = self.fem.computematrixDiffusion(self.kheatcell)
        lumped = True
        colorsrobin = bdrycond.colorsOfType("Robin")
        colorsdir = bdrycond.colorsOfType("Dirichlet")
        self.Arobin = self.fem.computeBdryMassMatrix(colorsrobin, bdrycond.param, lumped=self.masslumpedbdry)
        A += self.Arobin
        if 'convection' in self.problemdata.params.fct_glob.keys():
            A +=  self.fem.comptuteMatrixTransport(self.convection, self.convectionC, self.lamdconvection, colorsdir+colorsrobin)
        A, self.bdrydata = self.fem.matrixBoundary(A, method, bdrydata)
        return A
    def computeRhs(self, u=None):
        if not hasattr(self.bdrydata,"A_inner_dir"):
            raise ValueError("matrix() has to be called befor computeRhs()")
        bdrycond, method, bdrydata = self.problemdata.bdrycond, self.method, self.bdrydata
        b = np.zeros(self.fem.nunknowns())
        if 'rhs' in self.problemdata.params.fct_glob:
            fp1 = self.fem.interpolate(self.problemdata.params.fct_glob['rhs'])
            self.fem.massDot(b, fp1)
            if 'convection' in self.problemdata.params.fct_glob.keys():
                self.fem.massDotSupg(b, fp1, self.xd)
        if 'rhscell' in self.problemdata.params.fct_glob:
            fp1 = self.fem.interpolateCell(self.problemdata.params.fct_glob['rhscell'])
            self.fem.massDotCell(b, fp1)
        if 'rhspoint' in self.problemdata.params.fct_glob:
            self.fem.computeRhsPoint(b, self.problemdata.params.fct_glob['rhspoint'])
        colorsrobin = bdrycond.colorsOfType("Robin")
        colorsdir = bdrycond.colorsOfType("Dirichlet")
        colorsneu = bdrycond.colorsOfType("Neumann")
        if 'convection' in self.problemdata.params.fct_glob.keys():
            fp1 = self.fem.interpolateBoundary(colorsdir, bdrycond.fct)
            self.fem.massDotBoundary(b, fp1, coeff=-np.minimum(self.convection, 0), colors=colorsdir+colorsrobin)
        fp1 = self.fem.interpolateBoundary(colorsrobin, bdrycond.fct)
        self.fem.massDotBoundary(b, fp1, colorsrobin, lumped=self.masslumpedbdry, coeff=bdrycond.param)
        fp1 = self.fem.interpolateBoundary(colorsneu, bdrycond.fct)
        self.fem.massDotBoundary(b, fp1, colorsneu)
        b, u, self.bdrydata = self.fem.vectorBoundary(b, u, bdrycond, method, bdrydata)
        return b,u

    # def residualNewton(self, u):
    def initialCondition(self, expr):
        if not self._setMeshCalled: self.setMesh(self.mesh)
        self.Mass = self.fem.computeMassMatrix()
        fp1 = self.fem.interpolate(AnalyticalFunction(expr))
        b = np.zeros(self.fem.nunknowns())
        self.fem.massDot(b, fp1)
        u, niter = self.linearSolver(self.Mass, b)
        return u
    def dynamic(self, u0, niter=100, dt=0.1):
        if not hasattr(self,'Mass'):
            self.Mass = self.fem.computeMassMatrix()
        if not hasattr(self, 'Adyn'):
            self.A = self.computeMatrix()
            self.Adyn = self.Mass/dt + self.A
            # self.Adyn = self.Mass / dt + 0.5 * self.A
        u = u0.copy()
        for iter in range(niter):
            rhs,ub = self.computeRhs()
            rhs += (1/dt)*self.Mass.dot(u)
            # self.fem.massDot(rhs, u, coeff=1/dt)
            # rhs -= 0.5*self.A.dot(u)
            u, niter = self.linearSolver(self.Adyn, rhs)
        return u


    #     if not hasattr(self, 'du'): self.du = np.empty_like(u)
    #     self.du[:] = 0
    #     self.fem.formDiffusion(self.du, u, self.kheatcell)
    #     self.du -= self.b
    #     self.du = self.vectorDirichletZero(self.du, self.bdrydata)
    #     return self.du

    def postProcess(self, u):
        point_data, side_data, cell_data, global_data = {}, {}, {}, {}
        point_data['U'] = self.fem.tonode(u)
        if self.problemdata.solexact:
            global_data['error'] = {}
            global_data['error']['pcL2'], ec = self.fem.computeErrorL2Cell(self.problemdata.solexact, u)
            global_data['error']['pnL2'], en = self.fem.computeErrorL2Node(self.problemdata.solexact, u)
            global_data['error']['vcL2'] = self.fem.computeErrorFluxL2(self.problemdata.solexact, self.kheatcell, u)
            cell_data['E'] = ec
        global_data['postproc'] = {}
        if self.problemdata.postproc:
            types = ["bdry_mean", "bdry_fct", "bdry_nflux", "pointvalues", "meanvalues"]
            for name, type in self.problemdata.postproc.type.items():
                colors = self.problemdata.postproc.colors(name)
                if type == types[0]:
                    global_data['postproc'][name] = self.fem.computeBdryMean(u, colors)
                elif type == types[1]:
                    global_data['postproc'][name] = self.fem.computeBdryFct(u, colors)
                elif type == types[2]:
                    global_data['postproc'][name] = self.fem.computeBdryNormalFlux(u, colors, self.bdrydata, self.problemdata.bdrycond)
                elif type == types[3]:
                    global_data['postproc'][name] = self.fem.computePointValues(u, colors)
                elif type == types[4]:
                    global_data['postproc'][name] = self.fem.computeMeanValues(u, colors)
                else:
                    raise ValueError(f"unknown postprocess type '{type}' for key '{name}'\nknown types={types=}")
        if self.kheatcell.shape[0] != self.mesh.ncells:
            raise ValueError(f"self.kheatcell.shape[0]={self.kheatcell.shape[0]} but self.mesh.ncells={self.mesh.ncells}")
        cell_data['k'] = self.kheatcell
        return point_data, side_data, cell_data, global_data


#=================================================================#
if __name__ == '__main__':
    print("Pas de test")



import numpy as np
import scipy, scipy.special
import itertools as it

# ------------------------------------- #
def tensor(d, k):
    A = np.ones(shape=k*[d+1])
    facd = np.prod(np.arange(d + 1, d + k + 1))
    # print(f"{np.arange(d + 1, d + k + 1)=} {facd=}")
    for i in it.product(np.arange(d+1), repeat=k):
        A[i] = np.prod(scipy.special.factorial(np.bincount(i)))/facd
    return A

# ------------------------------------- #
if __name__ == '__main__':
    print(f"{tensor(d=3, k=2)=}")
    print(f"{tensor(d=3, k=1)=}")


# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sparse
try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh
import simfempy.fems.bdrydata


#=================================================================#
class FemCR1(object):
    def __init__(self, mesh=None):
        if mesh is not None:
            self.setMesh(mesh)
        self.dirichlet_al = 10

    def setMesh(self, mesh, bdrycond=None):
        self.mesh = mesh
        self.nloc = self.mesh.dimension+1
        self.cols = np.tile(self.mesh.facesOfCells, self.nloc).ravel()
        self.rows = np.repeat(self.mesh.facesOfCells, self.nloc).ravel()
        self.computeFemMatrices()
        self.massmatrix = self.massMatrix()
        if bdrycond:
            self.robinmassmatrix = self.computeBdryMassMatrix(bdrycond, type="Robin")

    def computeFemMatrices(self):
        ncells, normals, cellsOfFaces, facesOfCells, dV = self.mesh.ncells, self.mesh.normals, self.mesh.cellsOfFaces, self.mesh.facesOfCells, self.mesh.dV
        dim = self.mesh.dimension
        scale = 1
        self.cellgrads = scale*(normals[facesOfCells].T * self.mesh.sigma.T / dV.T).T
        scalemass = (2-dim) / (dim+1) / (dim+2)
        massloc = np.tile(scalemass, (self.nloc,self.nloc))
        massloc.reshape((self.nloc*self.nloc))[::self.nloc+1] = (2-dim + dim*dim) / (dim+1) / (dim+2)
        self.mass = np.einsum('n,kl->nkl', dV, massloc).ravel()

    def massMatrix(self):
        nfaces = self.mesh.nfaces
        self.massmatrix = sparse.coo_matrix((self.mass, (self.rows, self.cols)), shape=(nfaces, nfaces)).tocsr()
        return self.massmatrix

    # def computeRhs(self, u, rhs, kheatcell, bdrycond, method, bdrydata):
    def computeRhs(self, u, problemdata, kheatcell, method, bdrydata, robinmassmatrix):
        rhs = problemdata.rhs
        rhscell = problemdata.rhscell
        rhspoint = problemdata.rhspoint
        bdrycond = problemdata.bdrycond
        normals =  self.mesh.normals
        b = np.zeros(self.mesh.nfaces)
        if rhscell:
            x, y, z = self.mesh.pointsf.T
            b += self.massmatrix * rhscell(x, y, z)
        help = np.zeros(self.mesh.nfaces)
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != "Robin": continue
            xf, yf, zf = self.mesh.pointsf[faces].T
            nx, ny, nz = np.mean(normals[faces], axis=0)
            help[faces] = bdrycond.fct[color](xf, yf, zf, nx, ny, nz)
        b += robinmassmatrix*help
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != "Neumann": continue
            normalsS = normals[faces]
            dS = linalg.norm(normalsS,axis=1)
            normalsS = normalsS/dS[:,np.newaxis]
            xf, yf, zf = self.mesh.pointsf[faces].T
            nx, ny, nz = normalsS.T
            b[faces] += bdrycond.fct[color](xf, yf, zf, nx, ny, nz) * dS
        if bdrycond.hasExactSolution():
            for color, faces in self.mesh.bdrylabels.items():
                if bdrycond.type[color] != "Robin": continue
                normalsS = normals[faces]
                dS = linalg.norm(normalsS,axis=1)
                normalsS = normalsS/dS[:,np.newaxis]
                xf, yf, zf = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS.T
                b[faces] += bdrycond.fctexact["Robin"](xf, yf, zf, nx, ny, nz) * dS
        return self.vectorDirichlet(b, u, bdrycond, method, bdrydata)
    def matrixDiffusion(self, k, bdrycond, method, bdrydata):
        nfaces = self.mesh.nfaces
        matxx = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 0], self.cellgrads[:, :, 0])
        matyy = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 1], self.cellgrads[:, :, 1])
        matzz = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 2], self.cellgrads[:, :, 2])
        mat = ( (matxx+matyy+matzz).T*self.mesh.dV*k).T.ravel()
        A = sparse.coo_matrix((mat, (self.rows, self.cols)), shape=(nfaces, nfaces)).tocsr()
        return A
    def computeBdryMassMatrix(self, bdrycond, bdrycondtype, lumped):
        nfaces = self.mesh.nfaces
        rows = np.empty(shape=(0), dtype=int)
        cols = np.empty(shape=(0), dtype=int)
        mat = np.empty(shape=(0), dtype=float)
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != type: continue
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            cols = np.append(cols, faces)
            rows = np.append(rows, faces)
            mat = np.append(mat, bdrycond.param[color] * dS)
        return sparse.coo_matrix((mat, (rows, cols)), shape=(nfaces, nfaces)).tocsr()
    def prepareBoundary(self, colorsdir, postproc):
        bdrydata = simfempy.fems.bdrydata.BdryData()
        bdrydata.facesdirall = np.empty(shape=(0), dtype=int)
        bdrydata.colorsdir = colorsdir
        for color in colorsdir:
            facesdir = self.mesh.bdrylabels[color]
            bdrydata.facesdirall = np.unique(np.union1d(bdrydata.facesdirall, facesdir))
        bdrydata.facesinner = np.setdiff1d(np.arange(self.mesh.nfaces, dtype=int), bdrydata.facesdirall)
        bdrydata.facesdirflux = {}
        # for key, val in postproc.items():
        #     type, data = val.split(":")
        #     if type != "bdrydn": continue
        #     colors = [int(x) for x in data.split(',')]
        #     # bdrydata.facesdirflux[key] = np.empty(shape=(0), dtype=int)
        #     for color in colors:
        #         facesdir = self.mesh.bdrylabels[color]
        #         # bdrydata.facesdirflux[key] = np.unique(np.union1d(bdrydata.facesdirflux[key], facesdir).ravel())
        #         bdrydata.facesdirflux[color] = facesdir
        for name, type in postproc.type.items():
            if type != "bdrydn": continue
            colors = postproc.colors(name)
            for color in colors:
                facesdir = self.mesh.bdrylabels[color]
                bdrydata.facesdirflux[color] = facesdir
        return bdrydata
    def vectorDirichlet(self, b, u, bdrycond, method, bdrydata):
        facesdirflux, facesinner, facesdirall, colorsdir = bdrydata.facesdirflux, bdrydata.facesinner, bdrydata.facesdirall, bdrydata.colorsdir
        x, y, z = self.mesh.pointsf.T
        if u is None: u = np.zeros_like(b)
        else: assert u.shape == b.shape
        nfaces = self.mesh.nfaces
        # for key, faces in facesdirflux.items():
        #     bdrydata.bsaved[key] = b[faces]
        for color, faces in facesdirflux.items():
            bdrydata.bsaved[color] = b[faces]
        if method == 'trad':
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = bdrycond.fct[color]
                b[faces] = dirichlet(x[faces], y[faces], z[faces])
                u[faces] = b[faces]
            b[facesinner] -= bdrydata.A_inner_dir * b[facesdirall]
        else:
            for color in colorsdir:
                faces = self.mesh.bdrylabels[color]
                dirichlet = bdrycond.fct[color]
                u[faces] = dirichlet(x[faces], y[faces], z[faces])
                # b[faces] = 0
            b[facesinner] -= bdrydata.A_inner_dir * u[facesdirall]
            b[facesdirall] += bdrydata.A_dir_dir * u[facesdirall]
        return b, u, bdrydata
    def matrixDirichlet(self, A, bdrycond, method, bdrydata):
        facesdirflux, facesinner, facesdirall, colorsdir = bdrydata.facesdirflux, bdrydata.facesinner, bdrydata.facesdirall, bdrydata.colorsdir
        x, y, z = self.mesh.pointsf.T
        nfaces = self.mesh.nfaces
        # for key, faces in facesdirflux.items():
        #     nb = faces.shape[0]
        #     help = sparse.dok_matrix((nb, nfaces))
        #     for i in range(nb): help[i, faces[i]] = 1
        #     bdrydata.Asaved[key] = help.dot(A)
        for color, faces in facesdirflux.items():
            nb = faces.shape[0]
            help = sparse.dok_matrix((nb, nfaces))
            for i in range(nb): help[i, faces[i]] = 1
            bdrydata.Asaved[color] = help.dot(A)
        bdrydata.A_inner_dir = A[facesinner, :][:, facesdirall]
        if method == 'trad':
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help))
            help = np.zeros((nfaces))
            help[facesdirall] = 1.0
            help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            A += help
        else:
            bdrydata.A_dir_dir = self.dirichlet_al*A[facesdirall, :][:, facesdirall]
            help = np.ones((nfaces))
            help[facesdirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
            help2 = np.zeros((nfaces))
            help2[facesdirall] = np.sqrt(self.dirichlet_al)
            help2 = sparse.dia_matrix((help2, 0), shape=(nfaces, nfaces))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        return A, bdrydata
    def grad(self, ic):
        normals = self.mesh.normals[self.mesh.facesOfCells[ic,:]]
        grads = -normals/self.mesh.dV[ic]
        chsg =  (ic == self.mesh.cellsOfFaces[self.mesh.facesOfCells[ic,:],0])
        # print("### chsg", chsg, "normals", normals)
        grads[chsg] *= -1.
        return grads
    def computeErrorL2(self, solexact, uh):
        x, y, z = self.mesh.pointsf.T
        xc, yc, zc = self.mesh.pointsc.T
        e = solexact(x, y, z) - uh
        ec = solexact(xc, yc, zc) - np.mean(uh[self.mesh.facesOfCells], axis=1)
        return np.sqrt( np.dot(e, self.massmatrix*e) ), np.sqrt(np.sum(ec**2* self.mesh.dV)), e
    def computeErrorFluxL2(self, solexact, diffcell, uh):
        xc, yc, zc = self.mesh.pointsc.T
        graduh = np.einsum('nij,ni->nj', self.cellgrads, uh[self.mesh.facesOfCells])
        errv = 0
        for i in range(self.mesh.dimension):
            solxi = solexact.d(i, xc, yc, zc)
            errv += np.sum( diffcell*(solxi-graduh[:,i])**2* self.mesh.dV)
        return np.sqrt(errv)
    def computeBdryMean(self, u, colors):
        # colors = [int(x) for x in data.split(',')]
        mean, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            mean[i] = np.sum(dS*u[faces])
        return mean/omega
    def comuteFluxOnRobin(self, u, faces, dS, uR, cR):
        uhmean =  np.sum(dS * u[faces])
        xf, yf, zf = self.mesh.pointsf[faces].T
        normalsS = self.mesh.normals[faces]
        dS = linalg.norm(normalsS, axis=1)
        normalsS = normalsS / dS[:, np.newaxis]
        assert (dS.shape[0] == len(faces))
        nx, ny, nz = normalsS.T
        if uR: uRmean =  np.sum(dS * uR(xf, yf, zf, nx, ny, nz))
        else: uRmean=0
        return cR*(uRmean-uhmean)

    def computeBdryDn(self, u, colors, bdrydata, bdrycond):
        # colors = [int(x) for x in data.split(',')]
        flux, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            if bdrycond.type[color] == "Robin":
                flux[i] = self.comuteFluxOnRobin(u, faces, dS, bdrycond.fct[color], bdrycond.param[color])
            elif bdrycond.type[color] == "Dirichlet":
                bs, As = bdrydata.bsaved[color], bdrydata.Asaved[color]
                flux[i] = np.sum(As * u - bs)
            else:
                raise NotImplementedError("computeBdryDn for condition '{}'".format(bdrycond.type[color]))
        return flux

    def tonode(self, u):
        unodes = np.zeros(self.mesh.nnodes)
        scale = self.mesh.dimension
        np.add.at(unodes, self.mesh.simplices.T, np.sum(u[self.mesh.facesOfCells], axis=1))
        np.add.at(unodes, self.mesh.simplices.T, -scale*u[self.mesh.facesOfCells].T)
        countnodes = np.zeros(self.mesh.nnodes, dtype=int)
        np.add.at(countnodes, self.mesh.simplices.T, 1)
        unodes /= countnodes
        return unodes


#=================================================================#
if __name__ == '__main__':
    trimesh = SimplexMesh(geomname="backwardfacingstep", hmean=0.3)
    fem = FemCR1(trimesh)
    import plotmesh
    import matplotlib.pyplot as plt
    plotmesh.meshWithBoundaries(trimesh)
    plt.show()



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sparse
try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh
import simfempy.fems.bdrydata


#=================================================================#
class FemP1(object):
    def __init__(self, mesh=None):
        if mesh is not None:
            self.setMesh(mesh)
        self.dirichlet_al = 10

    def setMesh(self, mesh, bdrycond=None):
        self.mesh = mesh
        self.nloc = self.mesh.dimension+1
        simps = self.mesh.simplices
        self.cols = np.tile(simps, self.nloc).reshape(-1)
        self.rows = np.repeat(simps, self.nloc).reshape(-1)
        self.cellgrads = self.computeCellGrads()
        self.massmatrix = self.computeMassMatrix()
        if bdrycond:
            self.robinmassmatrix = self.computeBdryMassMatrix(bdrycond, type="Robin")

    def computeCellGrads(self):
        ncells, normals, cellsOfFaces, facesOfCells, dV = self.mesh.ncells, self.mesh.normals, self.mesh.cellsOfFaces, self.mesh.facesOfCells, self.mesh.dV
        scale = -1/self.mesh.dimension
        # print("dV", np.where(dV<0.0001))
        # print("dV", dV[dV<0.00001])
        return scale*(normals[facesOfCells].T * self.mesh.sigma.T / dV.T).T

    def computeMassMatrix(self, lumped=False):
        nnodes = self.mesh.nnodes
        scalemass = 1 / self.nloc / (self.nloc+1);
        massloc = np.tile(scalemass, (self.nloc,self.nloc))
        massloc.reshape((self.nloc*self.nloc))[::self.nloc+1] *= 2
        mass = np.einsum('n,kl->nkl', self.mesh.dV, massloc).ravel()
        return sparse.coo_matrix((mass, (self.rows, self.cols)), shape=(nnodes, nnodes)).tocsr()

    def computeBdryMassMatrix(self, bdrycond, type, lumped=False):
        # TODO: find a way to get linear solution exactly with lumped=True
        nnodes = self.mesh.nnodes
        rows = np.empty(shape=(0), dtype=int)
        cols = np.empty(shape=(0), dtype=int)
        mat = np.empty(shape=(0), dtype=float)
        if lumped:
            for color, faces in self.mesh.bdrylabels.items():
                if bdrycond.type[color] != type: continue
                scalemass = bdrycond.param[color]/ self.mesh.dimension
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                nodes = self.mesh.faces[faces]
                rows = np.append(rows, nodes)
                cols = np.append(cols, nodes)
                mass = np.repeat(scalemass*dS, self.mesh.dimension)
                mat = np.append(mat, mass)
            return sparse.coo_matrix((mat, (rows, cols)), shape=(nnodes, nnodes)).tocsr()
        else:
            for color, faces in self.mesh.bdrylabels.items():
                if bdrycond.type[color] != type: continue
                scalemass = bdrycond.param[color] / (1+self.mesh.dimension)/self.mesh.dimension
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                nodes = self.mesh.faces[faces]
                nloc = self.nloc-1
                rows = np.append(rows, np.repeat(nodes, nloc).reshape(-1))
                cols = np.append(cols, np.tile(nodes, nloc).reshape(-1))
                massloc = np.tile(scalemass, (nloc, nloc))
                massloc.reshape((nloc*nloc))[::nloc+1] *= 2
                mat = np.append(mat, np.einsum('n,kl->nkl', dS, massloc).reshape(-1))
            return sparse.coo_matrix((mat, (rows, cols)), shape=(nnodes, nnodes)).tocsr()

    def prepareBoundary(self, colorsdir, postproc):
        bdrydata = simfempy.fems.bdrydata.BdryData()
        bdrydata.nodesdir={}
        bdrydata.nodedirall = np.empty(shape=(0), dtype=int)
        for color in colorsdir:
            facesdir = self.mesh.bdrylabels[color]
            bdrydata.nodesdir[color] = np.unique(self.mesh.faces[facesdir].flat[:])
            bdrydata.nodedirall = np.unique(np.union1d(bdrydata.nodedirall, bdrydata.nodesdir[color]))
        bdrydata.nodesinner = np.setdiff1d(np.arange(self.mesh.nnodes, dtype=int),bdrydata.nodedirall)
        bdrydata.nodesdirflux={}
        if not postproc: return bdrydata
        for name, type in postproc.type.items():
            if type != "bdrydn": continue
            colors = postproc.colors(name)
            for color in colors:
                facesdir = self.mesh.bdrylabels[color]
                bdrydata.nodesdirflux[color] = np.unique(self.mesh.faces[facesdir].ravel())
        return bdrydata

    def matrixDiffusion(self, k, bdrycond, method, bdrydata):
        # alpha = problemdata.bdrycond.param[color]
        # print(f"??? {alpha=}")
        nnodes = self.mesh.nnodes
        matxx = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 0], self.cellgrads[:, :, 0])
        matyy = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 1], self.cellgrads[:, :, 1])
        matzz = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 2], self.cellgrads[:, :, 2])
        mat = ( (matxx+matyy+matzz).T*self.mesh.dV*k).T.ravel()
        A = sparse.coo_matrix((mat, (self.rows, self.cols)), shape=(nnodes, nnodes)).tocsr()
        A += self.robinmassmatrix
        return self.matrixDirichlet(A, bdrycond, method, bdrydata)

    def formDiffusion(self, du, u, k):
        graduh = np.einsum('nij,ni->nj', self.cellgrads, u[self.mesh.simplices])
        graduh = np.einsum('ni,n->ni', graduh, self.mesh.dV*k)
        # du += np.einsum('nj,nij->ni', graduh, self.cellgrads)
        raise ValueError(f"graduh {graduh.shape} {du.shape}")
        return du

    def computeRhs(self, u, problemdata, kheatcell, method, bdrydata):
        rhs = problemdata.rhs
        rhscell = problemdata.rhscell
        rhspoint = problemdata.rhspoint
        bdrycond = problemdata.bdrycond
        normals =  self.mesh.normals
        b = np.zeros(self.mesh.nnodes)
        if rhs:
            x, y, z = self.mesh.points.T
            b += self.massmatrix * rhs(x, y, z)
        if rhscell:
            scale = 1/(self.mesh.dimension+1)
            for label,fct in rhscell.items():
                if fct is None: continue
                cells = self.mesh.cellsoflabel[label]
                xc, yc, zc = self.mesh.pointsc[cells].T
                bC = scale*fct(xc, yc, zc)*self.mesh.dV[cells]
                # print("bC", bC)
                np.add.at(b, self.mesh.simplices[cells].T, bC)
        if rhspoint:
            for label,fct in rhspoint.items():
                if fct is None: continue
                points = self.mesh.verticesoflabel[label]
                xc, yc, zc = self.mesh.points[points].T
                # print("xc, yc, zc, f", xc, yc, zc, fct(xc, yc, zc))
                b[points] += fct(xc, yc, zc)

        help = np.zeros(self.mesh.nnodes)
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != "Robin": continue
            if not color in bdrycond.fct or bdrycond.fct[color] is None: continue
            nodes = np.unique(self.mesh.faces[faces].reshape(-1))
            x, y, z = self.mesh.points[nodes].T
            # print(f"normals {normals.shape}")
            # raise ValueError(f"normals = {np.mean(normals, axis=0)}")
            # nx, ny, nz = normals[faces].T
            nx, ny, nz = np.mean(normals[faces], axis=0)
            help[nodes] = bdrycond.fct[color](x, y, z, nx, ny, nz)
        b += self.robinmassmatrix*help

        scale = 1 / self.mesh.dimension
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != "Neumann": continue
            if not color in bdrycond.fct or bdrycond.fct[color] is None: continue
            normalsS = normals[faces]
            dS = linalg.norm(normalsS,axis=1)
            normalsS = normalsS/dS[:,np.newaxis]
            assert(dS.shape[0] == len(faces))
            x1, y1, z1 = self.mesh.pointsf[faces].T
            nx, ny, nz = normalsS.T
            bS = scale * bdrycond.fct[color](x1, y1, z1, nx, ny, nz) * dS
            np.add.at(b, self.mesh.faces[faces].T, bS)
        if bdrycond.hasExactSolution():
            for color, faces in self.mesh.bdrylabels.items():
                if bdrycond.type[color] != "Robin": continue
                normalsS = normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                normalsS = normalsS / dS[:, np.newaxis]
                assert (dS.shape[0] == len(faces))
                x1, y1, z1 = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS.T
                bS = scale * bdrycond.fctexact["Neumann"](x1, y1, z1, nx, ny, nz) * dS
                np.add.at(b, self.mesh.faces[faces].T, bS)
        return self.vectorDirichlet(b, u, bdrycond, method, bdrydata)

    def matrixDirichlet(self, A, bdrycond, method, bdrydata):
        nodesdir, nodedirall, nodesinner, nodesdirflux = bdrydata.nodesdir, bdrydata.nodedirall, bdrydata.nodesinner, bdrydata.nodesdirflux
        nnodes = self.mesh.nnodes
        for color, nodes in nodesdirflux.items():
            nb = nodes.shape[0]
            help = sparse.dok_matrix((nb, nnodes))
            for i in range(nb): help[i, nodes[i]] = 1
            bdrydata.Asaved[color] = help.dot(A)
        bdrydata.A_inner_dir = A[nodesinner, :][:, nodedirall]
        if method == 'trad':
            help = np.ones((nnodes))
            help[nodedirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nnodes, nnodes))
            A = help.dot(A.dot(help))
            help = np.zeros((nnodes))
            help[nodedirall] = 1.0
            help = sparse.dia_matrix((help, 0), shape=(nnodes, nnodes))
            A += help
        else:
            bdrydata.A_dir_dir = self.dirichlet_al*A[nodedirall, :][:, nodedirall]
            help = np.ones(nnodes)
            help[nodedirall] = 0
            help = sparse.dia_matrix((help, 0), shape=(nnodes, nnodes))
            help2 = np.zeros(nnodes)
            help2[nodedirall] = np.sqrt(self.dirichlet_al)
            help2 = sparse.dia_matrix((help2, 0), shape=(nnodes, nnodes))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        return A, bdrydata

    def vectorDirichlet(self, b, u, bdrycond, method, bdrydata):
        nodesdir, nodedirall, nodesinner, nodesdirflux = bdrydata.nodesdir, bdrydata.nodedirall, bdrydata.nodesinner, bdrydata.nodesdirflux
        if u is None: u = np.zeros_like(b)
        elif u.shape != b.shape : raise ValueError("u.shape != b.shape {} != {}".format(u.shape, b.shape))
        x, y, z = self.mesh.points.T
        for color, nodes in nodesdirflux.items():
            bdrydata.bsaved[color] = b[nodes]
        if method == 'trad':
            for color, nodes in nodesdir.items():
                if color in bdrycond.fct:
                    dirichlet = bdrycond.fct[color](x[nodes], y[nodes], z[nodes])
                    b[nodes] = dirichlet
                else:
                    b[nodes] = 0
                u[nodes] = b[nodes]
            b[nodesinner] -= bdrydata.A_inner_dir * b[nodedirall]
        else:
            for color, nodes in nodesdir.items():
                dirichlet = bdrycond.fct[color]
                if dirichlet:
                    u[nodes] = dirichlet(x[nodes], y[nodes], z[nodes])
                else:
                    u[nodes] = 0
                b[nodes] = 0
            b[nodesinner] -= bdrydata.A_inner_dir * u[nodedirall]
            b[nodedirall] += bdrydata.A_dir_dir * u[nodedirall]
        return b, u, bdrydata

    def vectorDirichletZero(self, du, bdrydata):
        nodesdir = bdrydata.nodesdir
        for color, nodes in nodesdir.items():
            du[nodes] = 0
        return du

    def tonode(self, u):
        return u

    # def grad(self, ic):
    #     normals = self.mesh.normals[self.mesh.facesOfCells[ic,:]]
    #     grads = 0.5*normals/self.mesh.dV[ic]
    #     chsg =  (ic == self.mesh.cellsOfFaces[self.mesh.facesOfCells[ic,:],0])
    #     # print("### chsg", chsg, "normals", normals)
    #     grads[chsg] *= -1.
    #     return grads

    def computeErrorL2(self, solexact, uh):
        x, y, z = self.mesh.points.T
        en = solexact(x, y, z) - uh
        xc, yc, zc = self.mesh.pointsc.T
        ec = solexact(xc, yc, zc) - np.mean(uh[self.mesh.simplices], axis=1)
        return np.sqrt( np.dot(en, self.massmatrix*en) ), np.sqrt(np.sum(ec**2* self.mesh.dV)), en

    def computeErrorFluxL2(self, solexact, diffcell, uh):
        xc, yc, zc = self.mesh.pointsc.T
        graduh = np.einsum('nij,ni->nj', self.cellgrads, uh[self.mesh.simplices])
        errv = 0
        for i in range(self.mesh.dimension):
            solxi = solexact.d(i, xc, yc, zc)
            errv += np.sum( diffcell*(solxi-graduh[:,i])**2* self.mesh.dV)
        return np.sqrt(errv)


    def computeBdryMean(self, u, colors):
        # colors = [int(x) for x in data.split(',')]
        mean, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            mean[i] = np.sum(dS*np.mean(u[self.mesh.faces[faces]],axis=1))
        return mean/omega

    def comuteFluxOnRobin(self, u, faces, dS, uR, cR):
        uhmean =  np.sum(dS * np.mean(u[self.mesh.faces[faces]], axis=1))
        xf, yf, zf = self.mesh.pointsf[faces].T
        nx, ny, nz = np.mean(self.mesh.normals[faces], axis=0)
        if uR: uRmean =  np.sum(dS * uR(xf, yf, zf, nx, ny, nz))
        else: uRmean=0
        return cR*(uRmean-uhmean)

    def computeBdryDn(self, u, colors, bdrydata, bdrycond):
        # colors = [int(x) for x in data.split(',')]
        flux, omega = np.zeros(len(colors)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            if bdrycond.type[color] == "Robin":
                flux[i] = self.comuteFluxOnRobin(u, faces, dS, bdrycond.fct[color], bdrycond.param[color])
            elif bdrycond.type[color] == "Dirichlet":
                bs, As = bdrydata.bsaved[color], bdrydata.Asaved[color]
                flux[i] = np.sum(As * u - bs)
            else:
                raise NotImplementedError(f"computeBdryDn for condition '{bdrycond.type[color]}' color={color}")
        return flux

    def computeBdryFct(self, u, colors):
        # colors = [int(x) for x in data.split(',')]
        nodes = np.empty(shape=(0), dtype=int)
        for color in colors:
            faces = self.mesh.bdrylabels[color]
            nodes = np.unique(np.union1d(nodes, self.mesh.faces[faces].ravel()))
        return self.mesh.points[nodes], u[nodes]

    def computePointValues(self, u, colors):
        # colors = [int(x) for x in data.split(',')]
        up = np.empty(len(colors))
        for i,color in enumerate(colors):
            nodes = self.mesh.verticesoflabel[color]
            up[i] = u[nodes]
        return up

    def computeMeanValues(self, u, colors):
        # colors = [int(x) for x in data.split(',')]
        up = np.empty(len(colors))
        for i, color in enumerate(colors):
            up[i] = self.computeMeanValue(u,color)
        return up

    def computeMeanValue(self, u, color):
        cells = self.mesh.cellsoflabel[color]
        # print("umean", np.mean(u[self.mesh.simplices[cells]],axis=1))
        return np.sum(np.mean(u[self.mesh.simplices[cells]],axis=1)*self.mesh.dV[cells])


# ------------------------------------- #

if __name__ == '__main__':
    trimesh = SimplexMesh(geomname="backwardfacingstep", hmean=0.3)
    fem = FemP1(trimesh)
    fem.testgrad()
    import plotmesh
    import matplotlib.pyplot as plt
    plotmesh.meshWithBoundaries(trimesh)
    plt.show()



# just to make sure the local simfempy is found first
from os import path
import sys
simfempypath = path.dirname(path.dirname(path.abspath(__file__)))
sys.path.insert(0,simfempypath)
# just to make sure the local simfempy is found first


import pygmsh
import numpy as np
import simfempy
import matplotlib.pyplot as plt


# ------------------------------------- #
def pygmshexample():
    with pygmsh.geo.Geometry() as geom:
        poly = geom.add_polygon(
            [
                [+0.0, +0.5],
                [-0.1, +0.1],
                [-0.5, +0.0],
                [-0.1, -0.1],
                [+0.0, -0.5],
                [+0.1, -0.1],
                [+0.5, +0.0],
                [+0.1, +0.1],
            ],
            mesh_size=0.05,
        )
        geom.add_physical(poly.surface, label="100")
        top, vol, ext = geom.twist(
            poly,
            translation_axis=[0, 0, 1],
            rotation_axis=[0, 0, 1],
            point_on_axis=[0, 0, 0],
            angle=np.pi / 3,
        )
        geom.add_physical(top, label=f"{101+len(ext)}")
        for i in range(len(ext)): geom.add_physical(ext[i], label=f"{101+i}")
        geom.add_physical(vol, label="10")
        mesh = geom.generate_mesh()
    return mesh


def createData(bdrylabels):
    data = simfempy.applications.problemdata.ProblemData()
    bdrylabels = list(bdrylabels)
    labels_lat = bdrylabels[1:-1]
    firstlabel = bdrylabels[0]
    lastlabel = bdrylabels[-1]
    labels_td = [firstlabel,lastlabel]
    bdrycond =  data.bdrycond
    bdrycond.set("Neumann", labels_lat)
    bdrycond.set("Dirichlet", labels_td)
    bdrycond.fct[firstlabel] = lambda x,y,z: 200
    bdrycond.fct[lastlabel] = lambda x,y,z: 100
    postproc = data.postproc
    postproc.set(name="lateral mean", type='bdry_mean', colors=labels_lat)
    postproc.set(name="normal fux", type='bdry_nflux', colors=labels_td)
    # postproc.type['bdry_mean'] = "bdry_mean"
    # postproc.color['bdry_mean'] = labels_lat
    # postproc.type['bdry_nflux'] = "bdry_nflux"
    # postproc.color['bdry_nflux'] = labels_td
    data.params.scal_glob["kheat"] = 0.0001
    return data

# ------------------------------------- #
mesh = pygmshexample()
mesh = simfempy.meshes.simplexmesh.SimplexMesh(mesh=mesh)
# simfempy.meshes.plotmesh.meshWithBoundaries(mesh)

data = createData(mesh.bdrylabels.keys())
print("data", data)
data.check(mesh)

heat = simfempy.applications.heat.Heat(problemdata=data, mesh=mesh)
result = heat.static()
print(f"{result.info['timer']}")
print(f"{result.info['iter']}")
print(f"postproc: {result.data['global']['postproc']}")
simfempy.meshes.plotmesh.meshWithData(mesh, data=result.data)
plt.show()



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sparse
try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh
import simfempy.fems.bdrydata

#=================================================================#
class FemRT0(object):
    """
    on suppose que  self.mesh.edgesOfCell[ic, kk] et oppose à elem[ic,kk] !!!
    """
    def __init__(self, mesh=None, massproj=None):
        if mesh is not None:
            self.setMesh(mesh)
        self.massproj=massproj

    def setMesh(self, mesh):
        self.mesh = mesh
        self.nloc = self.mesh.dimension+1
        self.Mtocell = self.toCellMatrix()

    def toCellMatrix(self):
        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        dim, dV, nloc, p, pc, simp = self.mesh.dimension, self.mesh.dV, self.nloc, self.mesh.points, self.mesh.pointsc, self.mesh.simplices
        dS = sigma * linalg.norm(normals[facesofcells], axis=2)/dim
        ps = p[simp][:,:,:dim]
        ps2 = np.transpose(ps, axes=(2,0,1))
        pc2 = np.repeat(pc[:,:dim].T[:, :, np.newaxis], nloc, axis=2)
        pd = pc2 -ps2
        rows = np.repeat((np.repeat(dim * np.arange(ncells), dim).reshape(ncells,dim) + np.arange(dim)).swapaxes(1,0),nloc)
        cols = np.tile(facesofcells.ravel(), dim)
        mat = np.einsum('ni, jni, n->jni', dS, pd, 1/dV)
        return  sparse.coo_matrix((mat.ravel(), (rows.ravel(), cols.ravel())), shape=(dim*ncells, nfaces))

    def toCell(self, v):
        return self.Mtocell.dot(v)

    def constructMass(self, diffinvcell=None):
        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        dim, dV, nloc, simp = self.mesh.dimension, self.mesh.dV, self.nloc, self.mesh.simplices
        p, pc, pf = self.mesh.points, self.mesh.pointsc, self.mesh.pointsf


        if self.massproj is None:
            # RT
            scalea = 1 / dim / dim / (dim + 2) / (dim + 1)
            scaleb = 1 / dim / dim / (dim + 2) * (dim + 1)
            scalec = 1 / dim / dim
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            x1 = scalea *np.einsum('nij,nij->n', p[simp], p[simp]) + scaleb* np.einsum('ni,ni->n', pc, pc)
            x2 = scalec *np.einsum('nik,njk->nij', p[simp], p[simp])
            x3 = - scalec * np.einsum('nik,nk->ni', p[simp], pc)
            mat = np.einsum('ni,nj, n->nij', dS, dS, x1)
            mat += np.einsum('ni,nj,nij->nij', dS, dS, x2)
            mat += np.einsum('ni,nj,ni->nij', dS, dS, x3)
            mat += np.einsum('ni,nj,nj->nij', dS, dS, x3)
            if diffinvcell is None:
                mat = np.einsum("nij, n -> nij", mat, 1/dV)
            else:
                mat = np.einsum("nij, n -> nij", mat, diffinvcell / dV  )
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces)).tocsr()
            # print("A (RT)", A)
            return A

        elif self.massproj=="L2":
            # RT avec projection L2
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)/dim
            ps = p[simp][:,:,:dim]
            ps2 = np.transpose(ps, axes=(2,0,1))
            pc2 = np.repeat(pc[:,:dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 -ps2
            mat = np.einsum('kni,knj, ni, nj, n->nij', pd, pd, dS, dS, diffinvcell / dV)
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces)).tocsr()
            # print("A (RTM)", A)
            return A


        elif self.massproj == "RT_Bar":
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            scale = 1/ (dim+1)
            scale = 2/9
            mat = np.einsum('ni, nj, n->nij', -dS, 1/dS, dV)
            mat.reshape( ( mat.shape[0], (dim+1)**2) ) [:,::dim+2] *= -dim
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.tocsr()
        elif self.massproj == "Bar_RT":
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            scale = 1/ (dim+1)
            scale = 2/9
            mat = np.einsum('ni, nj, n->nij', -dS, 1/dS, dV)
            mat.reshape( ( mat.shape[0], (dim+1)**2) ) [:,::dim+2] *= -dim
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.tocsr().T

        elif self.massproj == "Hat_RT":
            # PG de type RT-Hat (Hat aligned with "m")
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            scale = 1 / dim / dim
            mat = np.einsum('kni, knj, ni, nj, n->nij', pd, pd, dS, dS, 1 / dV)
            # pas la si projection L2
            # mat += np.einsum('kni, kni, ni, nj, n->nij', pd, pd, dS, dS, dim / (dim + 2) / dV)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.tocsr().T

        elif self.massproj == "Hat_Hat":
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            mloc = np.tile(2-dim, (dim+1, dim+1))
            mloc.reshape(( (dim+1)**2))[::dim+2] += dim*dim
            scale = (dim+1) / (dim+2) / dim**2
            mat = np.einsum('kni, knj, ij, ni, nj, n->nij', pd, pd, mloc, dS, dS, 1 / dV)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.tocsr()

        elif self.massproj=="RT_Tilde":
            # PG de type RT-Tilde (Hat aligned with "n")
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            dT = 1/linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            pn = np.transpose(normals[facesofcells][:,:,:dim], axes=(2,0,1))
            # multiplié par dim !
            scale = dim / dim / (dim+1)
            mat = np.einsum('kni, knj, ni, nj, n->nij', pn, pd, dT, dS, diffinvcell)
            mat += np.einsum('kni, kni, ni, nj, n->nij', pn, pd, dT, dS, dim/(dim+2) *diffinvcell)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces)).tocsr()
            # A[np.abs(A)<1e-10] = 0
            # A.eliminate_zeros()
            # print("A (RTxTilde)", A)
            return A

        elif self.massproj=="Tilde_RT":
            # PG de type RT-Tilde (Hat aligned with "n")
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            dT = 1/linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            pn = np.transpose(normals[facesofcells][:,:,:dim], axes=(2,0,1))
            # multiplié par d !
            scale = dim / dim / (dim+1)
            mat = np.einsum('kni, knj, ni, nj, n->nji', pn, pd, dT, dS, diffinvcell)
            mat += np.einsum('kni, kni, ni, nj, n->nji', pn, pd, dT, dS, dim/(dim+2) *diffinvcell)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.tocsr().T

        elif self.massproj=="HatxRTOLD":
            # PG de type Tilde-RT
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            pf2 = pf[facesofcells][:, :, :dim]
            scale = 1 / dim / dim
            mat = np.einsum('kni, nik, nj, ni, n->nij', pd, pf2, dS, dS, 1 / dV)
            mat -= np.einsum('kni, njk, nj, ni, n->nij', pd, ps, dS, dS, 1 / dV)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces)).tocsr()
            # print("A (HatxRT)", A)
            return A

        elif self.massproj=="RTxHatOLD":
            # PG de type RT-Hat (Hat aligned with "m")
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            pf2 = pf[facesofcells][:, :, :dim]
            scale = 1 / dim / dim
            mat = np.einsum('kni, nik, nj, ni, n->nij', pd, pf2, dS, dS, 1 / dV)
            mat -= np.einsum('kni, njk, nj, ni, n->nij', pd, ps, dS, dS, 1 / dV)
            mat *= scale
            rows = np.repeat(facesofcells, self.nloc).ravel()
            cols = np.tile(facesofcells, self.nloc).ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces))
            return A.T.tocsr()

        elif self.massproj=="HatxHatOLD":
            # G de type Tilde-Tilde
            dS = sigma * linalg.norm(normals[facesofcells], axis=2)
            ps = p[simp][:, :, :dim]
            ps2 = np.transpose(ps, axes=(2, 0, 1))
            pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
            pd = pc2 - ps2
            scale = (dim + 1) / dim**3
            mat = scale * np.einsum('ni, ni, kni, kni, n->ni', dS, dS, pd, pd, diffinvcell / dV)
            rows = facesofcells.ravel()
            A = sparse.coo_matrix((mat.ravel(), (rows, rows)), shape=(nfaces, nfaces)).tocsr()
            # print("A", A)
            return A

        else:
            raise ValueError("unknown type self.massproj={}".format(self.massproj))

    def constructDiv(self):
        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        rows = np.repeat(np.arange(ncells), self.nloc)
        cols = facesofcells.ravel()
        mat =  (sigma*linalg.norm(normals[facesofcells],axis=2)).ravel()
        return  sparse.coo_matrix((mat, (rows, cols)), shape=(ncells, nfaces)).tocsr()

    def reconstruct(self, p, vc, diffinv):
        nnodes, ncells, dim = self.mesh.nnodes, self.mesh.ncells, self.mesh.dimension
        if len(diffinv.shape) != 1:
            raise NotImplemented("only scalar diffusion the time being")
        counts = np.bincount(self.mesh.simplices.reshape(-1))
        pn2 = np.zeros(nnodes)
        xdiff = self.mesh.points[self.mesh.simplices, :dim] - self.mesh.pointsc[:, np.newaxis,:dim]
        rows = np.repeat(self.mesh.simplices,dim)
        cols = np.repeat(dim*np.arange(ncells),dim*(dim+1)).reshape(ncells * (dim+1), dim) + np.arange(dim)
        mat = np.einsum("nij, n -> nij", xdiff, diffinv)
        A = sparse.coo_matrix((mat.reshape(-1), (rows.reshape(-1), cols.reshape(-1))), shape=(nnodes, dim*ncells)).tocsr()
        np.add.at(pn2, self.mesh.simplices.T, p)
        pn2 += A*vc
        pn2 /= counts
        return pn2


    def rhsDirichlet(self, faces, ud):
        return linalg.norm(self.mesh.normals[faces],axis=1) * ud

    def constructRobin(self, bdrycond, type):
        nfaces = self.mesh.nfaces
        rows = np.empty(shape=(0), dtype=int)
        cols = np.empty(shape=(0), dtype=int)
        mat = np.empty(shape=(0), dtype=float)
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != type: continue
            if not bdrycond.param[color]: continue
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            cols = np.append(cols, faces)
            rows = np.append(rows, faces)
            mat = np.append(mat, 1/bdrycond.param[color] * dS)
        A = sparse.coo_matrix((mat, (rows, cols)), shape=(nfaces, nfaces)).tocsr()
        return A

    def matrixNeumann(self, A, B, bdrycond):
        nfaces = self.mesh.nfaces
        bdrydata = simfempy.fems.bdrydata.BdryData()
        bdrydata.facesneumann = np.empty(shape=(0), dtype=int)
        bdrydata.colorsneum = bdrycond.colorsOfType("Neumann")
        for color in bdrydata.colorsneum:
            bdrydata.facesneumann = np.unique(np.union1d(bdrydata.facesneumann, self.mesh.bdrylabels[color]))
        bdrydata.facesinner = np.setdiff1d(np.arange(self.mesh.nfaces, dtype=int), bdrydata.facesneumann)

        bdrydata.B_inner_neum = B[:, :][:, bdrydata.facesneumann]
        help = np.ones(nfaces)
        help[bdrydata.facesneumann] = 0
        help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
        B = B.dot(help)

        bdrydata.A_inner_neum = A[bdrydata.facesinner, :][:, bdrydata.facesneumann]
        bdrydata.A_neum_neum = A[bdrydata.facesneumann, :][:, bdrydata.facesneumann]
        help2 = np.zeros((nfaces))
        help2[bdrydata.facesneumann] = 1
        help2 = sparse.dia_matrix((help2, 0), shape=(nfaces, nfaces))
        A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))

        return bdrydata, A, B



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from mpl_toolkits import mplot3d

#----------------------------------------------------------------#
def _settitle(ax, text):
    try:
        ax.set_title(text)
    except:
        ax.title(text)

#----------------------------------------------------------------#
def _plotNodeLabels(x, y, z, ax=plt):
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
    for i in range(len(x)):
        ax.text(x[i], y[i], z[i], r'%d' % (i), fontweight='bold', bbox=props)


#=================================================================#
def _plotCells(x, y, z, tets, ax=plt):
    for i in range(len(tets)):
        ax.plot(x[tets[i]], y[tets[i]], z[tets[i]], color=(0.5,0.5,0.5))

#=================================================================#
def _plotCellLabels(tets, xc, yc, zc, ax=plt):
    for i in range(len(tets)):
        ax.text(xc[i], yc[i], zc[i], r'%d' % (i), color='r', fontweight='bold', fontsize=10)


#=================================================================#
def meshWithBoundaries(x, y, z, tets, faces, bdrylabels, nodelabels=False, ax=plt):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel(r'x')
    ax.set_ylabel(r'y')
    ax.set_zlabel(r'z')
    # if nodelabels: _plotNodeLabels(x, y, z, ax=ax)
    _plotCells(x, y, z, tets, ax=ax)
    cmap = plt.get_cmap("tab10")
    patches=[]
    i=0
    for color, bdryfaces in bdrylabels.items():
        patches.append(mpatches.Patch(color=cmap(i), label=color))
        for ie in bdryfaces:
            poly3d = [ [x[f], y[f], z[f]] for f in faces[ie]]
            ax.add_collection3d(Poly3DCollection([poly3d], facecolors=cmap(i), linewidths=1))
        i += 1
    ax.legend(handles=patches)
    _settitle(ax, "Mesh and Boundary Labels")

#=================================================================#
def plotmesh(mesh, **kwargs):
    import pyvista
    import vtk
    if 'ax' in kwargs: ax = kwargs.pop('ax')
    else: ax = plt
    title = 'Mesh'
    if 'title' in kwargs: title = kwargs.pop('title')
    x, y, z, tets = mesh.points[:, 0], mesh.points[:, 1], mesh.points[:, 2], mesh.simplices

    xyz = np.stack((x, y, z)).T
    ntets = tets.shape[0]
    cell_type = vtk.VTK_TETRA*np.ones(ntets, dtype=int)
    offset = 5*np.arange(ntets)
    cells = np.insert(tets, 0, 4, axis=1).ravel()
    grid = pyvista.UnstructuredGrid(offset, cells, cell_type, xyz)

    plotter = pyvista.Plotter()
    plotter.renderer.SetBackground(255, 255, 255)
    plotter.add_axes()
    plotter.add_mesh(grid, showedges=False, opacity=0.6, color='gray')
    plotter.remove_scalar_bar()
    pltcolors = 'bgrcmykbgrcmyk'
    for i, (color, vertices) in enumerate(mesh.verticesoflabel.items()):
        # print("color", color, "vertices", vertices)
        points = mesh.points[vertices]
        nv = vertices.shape[0]
        # plotter.add_points(points, scalars=i*np.ones(nv), cmap=None, flip_scalars=True)
        plotter.add_points(points, color=pltcolors[i], cmap=None, flip_scalars=True, point_size=15)
        # plotter.add_text('x', points[0], color=pltcolors[i])
        # patches.append(mpatches.Patch(color=pltcolors[i], label=color))
        # for vertex in vertices:
        #     ax.plot(x[vertex], y[vertex],'X', color=pltcolors[i])
    cpos = plotter.show(title="U")

    # ax.triplot(x, y, tris, color='k')
    # if ax ==plt:
    #     plt.gca().set_aspect(aspect='equal')
    #     ax.xlabel(r'x')
    #     ax.ylabel(r'y')
    # else:
    #     ax.set_aspect(aspect='equal')
    #     ax.set_xlabel(r'x')
    #     ax.set_ylabel(r'y')
    # celllabels = mesh.cell_labels
    # cnt = ax.tripcolor(x, y, tris, facecolors=celllabels, edgecolors='k', cmap='jet', alpha=0.4)
    # # clb = plt.colorbar(cnt)
    # # clb.set_label("cellcolors")
    # pltcolors = 'bgrcmykbgrcmyk'
    # patches=[]
    # for i, (color, vertices) in enumerate(mesh.verticesoflabel.items()):
    #     patches.append(mpatches.Patch(color=pltcolors[i], label=color))
    #     for vertex in vertices:
    #         ax.plot(x[vertex], y[vertex],'X', color=pltcolors[i])
    # ax.legend(handles=patches)
    # _settitle(ax, title)

# =================================================================#
def meshWithData(**kwargs):
    import pyvista
    import vtk

    x, y, z, tets = kwargs['x'], kwargs['y'], kwargs['z'], kwargs['tets']
    addplots = []
    if 'addplots' in kwargs: addplots = kwargs['addplots']
    if addplots is None: addplots=[]
    point_data, cell_data, quiver_cell_data, translate_point_data = None, None, None, None
    title, suptitle = None, None
    if 'data' in kwargs:
        point_data = kwargs['data']['point']
        cell_data = kwargs['data']['cell']
    # if 'point_data' in kwargs: point_data = kwargs['point_data']
    # if 'cell_data' in kwargs: cell_data = kwargs['cell_data']
    if 'quiver_cell_data' in kwargs: quiver_cell_data = kwargs['quiver_cell_data']
    if 'translate_point_data' in kwargs: translate_point_data = kwargs['translate_point_data']
    if 'numbering' in kwargs: numbering = kwargs['numbering']
    if 'title' in kwargs: title = kwargs['title']
    if 'suptitle' in kwargs: suptitle = kwargs['suptitle']

    xyz = np.stack((x, y, z)).T
    ntets = tets.shape[0]
    cell_type = vtk.VTK_TETRA*np.ones(ntets, dtype=int)
    cells = np.insert(tets, 0, 4, axis=1).ravel()
    grid = pyvista.UnstructuredGrid(cells, cell_type, xyz)
    # offset = 5*np.arange(ntets)
    # grid = pyvista.UnstructuredGrid(offset, cells, cell_type, xyz)

    if translate_point_data:
        assert len(point_data.keys()) ==3
        u, unames = [], []
        for k,v in point_data.items():
            unames.append(k)
            u.append(v)
        scale = translate_point_data
        un = np.sqrt(u[0]**2+u[1]**2+u[2]**2)
        xyz2 = np.stack((x+scale*u[0], y+scale*u[1], z+scale*u[2])).T
        grid2 = pyvista.UnstructuredGrid(offset, cells, cell_type, xyz2)
        plotter = pyvista.Plotter()
        plotter.renderer.SetBackground(255,255,255)
        plotter.add_axes()
        # plotter.add_mesh(grid, stitle="U=0", showedges=False, opacity=0.6, color='gray')
        plotter.add_mesh(grid, stitle="U=0", opacity=0.6, color='gray')
        plotter.remove_scalar_bar()
        # plotter.add_mesh(grid2, scalars=un, stitle="U", showedges=True, interpolatebeforemap=True)
        plotter.add_mesh(grid2, scalars=un, stitle="U")
        plotter.remove_scalar_bar()
        cpos = plotter.show(title="U")
    else:
        plotter = pyvista.Plotter()
        count=0
        for pdn, pd in point_data.items():
            # grid.plot(scalars=pd, stitle=pdn)
            # plotter.add_mesh(grid, scalars=pd, stitle=pdn,showedges=True,interpolatebeforemap=True)
            plotter.add_mesh(grid, scalars=pd, stitle=pdn)
            count += 1
        cpos = plotter.show()
    return

# =================================================================#
def meshWithData2(x, y, z, tets, xc, yc, zc, point_data, cell_data, ax=plt, numbering=False, title=None, suptitle=None,addplots=[]):
    nplots = len(point_data)+len(cell_data)
    if nplots==0:
        print("meshWithData() no point_data")
        return
    fig = plt.figure(figsize=plt.figaspect(1/nplots))
    if suptitle: fig.suptitle(suptitle)
    axs = []
    for i in range(nplots):
        axs.append(fig.add_subplot("1{:1d}{:1d}".format(nplots, i+1), projection='3d'))
    count=0
    for pdn, pd in point_data.items():
        ax = axs[count]
        xyz = np.stack((x,y,z)).T
        print("xyz", xyz.shape)
        vts = xyz[tets, :]
        tri = mplot3d.art3d.Poly3DCollection(vts)
        tri.set_alpha(0.2)
        tri.set_color('grey')
        ax.add_collection3d(tri)
        ax.plot(x,y,z, 'k.')
        ax.set_axis_off()
        # for i in range(len(tets)):
        #     ax.plot(x[tets[i]], y[tets[i]], z[tets[i]])
            # ax.plot_trisurf(x[tets[i]], y[tets[i]], z[tets[i]])
        _settitle(ax, pdn)
        count += 1
    if title: fig.canvas.set_window_title(title)
    plt.show()

#=================================================================#
def meshWithData2(**kwargs):
    import pyvista
    import vtk

    import matplotlib
    matplotlib.use('Agg')
    from matplotlib.figure import Figure
    from matplotlib.backends.backend_agg import FigureCanvasAgg
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = fig.add_subplot(111)
    ax.grid(True)
    ax.set_xlabel('Hello from VTK!', size=16)
    ax.bar(xrange(10), p.rand(10))
    # The pyvistamageImporter will treat a python string as a void pointer
    importer = pyvistamageImport()
    importer.SetDataScalarTypeToUnsignedChar()
    importer.SetNumberOfScalarComponents(4)

    # It's upside-down when loaded, so add a flip filter
    imflip = pyvistamageFlip()
    imflip.SetInput(importer.GetOutput())
    imflip.SetFilteredAxis(1)

    # Map the plot as a texture on a cube
    cube = vtkCubeSource()

    cubeMapper = vtkPolyDataMapper()
    cubeMapper.SetInput(cube.GetOutput())

    cubeActor = vtkActor()
    cubeActor.SetMapper(cubeMapper)

    # Create a texture based off of the image
    cubeTexture = vtkTexture()
    cubeTexture.InterpolateOn()
    cubeTexture.SetInput(imflip.GetOutput())
    cubeActor.SetTexture(cubeTexture)

    ren = vtkRenderer()
    ren.AddActor(cubeActor)

    renWin = vtkRenderWindow()
    renWin.AddRenderer(ren)

    iren = vtkRenderWindowInteractor()
    iren.SetRenderWindow(renWin)
    # Powers of 2 image to be clean
    w, h = 1024, 1024
    dpi = canvas.figure.get_dpi()
    fig.set_figsize_inches(w / dpi, h / dpi)
    canvas.draw()  # force a draw

    # This is where we tell the image importer about the mpl image
    extent = (0, w - 1, 0, h - 1, 0, 0)
    importer.SetWholeExtent(extent)
    importer.SetDataExtent(extent)
    importer.SetImportVoidPointer(canvas.buffer_rgba(0, 0), 1)
    importer.Update()

    iren.Initialize()
    iren.Start()




import numpy as np
from simfempy import solvers
from simfempy import fems

#=================================================================#
class Elliptic(solvers.solver.Application):
    """
    """
    def defineRhsAnalyticalSolution(self, solexact):
        def _fctu0(x, y, z, diff):
            rhs = np.zeros(x.shape[0])
            for i in range(self.mesh.dimension):
                rhs -= diff * solexact[0].dd(i, i, x, y, z)
            return rhs
        def _fctu1(x, y, z, diff):
            rhs = np.zeros(x.shape[0])
            for i in range(self.mesh.dimension):
                rhs -= diff * solexact[1].dd(i, i, x, y, z)
            return rhs
        return [_fctu0, _fctu1]

    def defineNeumannAnalyticalSolution_0(self, solexact):
        def _fctneumann(x, y, z, nx, ny, nz, diff):
            rhs = np.zeros(x.shape[0])
            normals = nx, ny, nz
            for i in range(self.mesh.dimension):
                rhs += diff * solexact[0].d(i, x, y, z) * normals[i]
            return rhs
        return _fctneumann

    def defineNeumannAnalyticalSolution_1(self, solexact):
        def _fctneumann(x, y, z, nx, ny, nz, diff):
            rhs = np.zeros(x.shape[0])
            normals = nx, ny, nz
            for i in range(self.mesh.dimension):
                rhs += diff * solexact[1].d(i, x, y, z) * normals[i]
            return rhs
        return _fctneumann

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'geometry' in kwargs:
            return
        self.linearsolver = 'pyamg'
        if 'fem' in kwargs: fem = kwargs.pop('fem')
        else: fem='p1'
        if fem == 'p1':
            self.fem = fems.femp1sys.FemP1()
        elif fem == 'cr1':
            self.fem = fems.femcr1sys.FemCR1()
        else:
            raise ValueError("unknown fem '{}'".format(fem))
        if self.problemdata.postproc:
            print("self.problemdata.postproc", self.problemdata.postproc, len(self.problemdata.postproc), self.ncomp)
            assert len(self.problemdata.postproc) == self.ncomp
        if 'diff' in kwargs:
            self.diff = kwargs.pop('diff')
            assert len(self.diff) == self.ncomp
            for i in range(self.ncomp):
                self.diff[i] = np.vectorize(self.diff[i])
        else:
            self.diff = []
            for i in range(self.ncomp):
                self.diff.append(np.vectorize(lambda j: 0.123*(i+1)))
        if 'method' in kwargs: self.method = kwargs.pop('method')
        else: self.method="trad"
        if 'show_diff' in kwargs: self.show_diff = kwargs.pop('show_diff')
        else: self.show_diff=False

    def setMesh(self, mesh):
        super().setMesh(mesh)
        self.fem.setMesh(self.mesh, self.ncomp)
        self.diffcell = []
        for icomp in range(self.ncomp):
            self.diffcell.append(self.diff[icomp](self.mesh.cell_labels))
        self.bdrydata = self.fem.prepareBoundary(self.problemdata.bdrycond, self.problemdata.postproc)

    def solvestatic(self):
        return self.solveLinear()
        
    def solve(self, iter, dirname):
        return self.solveLinear()

    def computeRhs(self, u=None):
        b, u, self.bdrydata = self.fem.computeRhs(u, self.problemdata.rhs, self.diffcell, self.problemdata.bdrycond, self.method, self.bdrydata)
        return b,u

    def matrix(self):
        A, self.bdrydata = self.fem.matrixDiffusion(self.diffcell, self.problemdata.bdrycond, self.method, self.bdrydata)
        return A

    def postProcess(self, u):
        info = {}
        cell_data = {}
        point_data = {}
        for icomp in range(self.ncomp):
            point_data['U_{:02d}'.format(icomp)] = self.fem.tonode(u[icomp::self.ncomp])
        if self.problemdata.solexact:
            info['error'] = {}
            err, e = self.fem.computeErrorL2(self.problemdata.solexact, u)
            info['error']['L2'] = np.sum(err)
            for icomp in range(self.ncomp):
                point_data['E_{:02d}'.format(icomp)] = self.fem.tonode(e[icomp])
        info['postproc'] = {}
        print("self.problemdata.postproc", self.problemdata.postproc)
        for icomp, postproc in enumerate(self.problemdata.postproc):
            for key, val in postproc.items():
                type,data = val.split(":")
                if type == "bdrymean":
                    info['postproc']["{}_{:02d}".format(key,icomp)] = self.fem.computeBdryMean(u, data, icomp)
                elif type == "bdrydn":
                    info['postproc']["{}_{:02d}".format(key,icomp)] = self.fem.computeBdryDn(u, data, self.bdrydata, self.problemdata.bdrycond, icomp)
                else:
                    raise ValueError("unknown postprocess {}".format(key))
            if self.show_diff: cell_data['diff_{:02d}'.format(icomp)] = self.diffcell[icomp]
        return point_data, cell_data, info

#=================================================================#
if __name__ == '__main__':
    print("Pas encore de test")



from . import bdrydata, p1, rt0
from . import femd0
from . import femp1
from . import femp1sys
from . import femcr1
from . import femcr1sys
from . import femrt0
from . import fembv0
from . import fembv0
